# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.observation._generated_parser import (
    UnexpectedInput,
    Lark_StandAlone,
)

PARSER = Lark_StandAlone()


class TestParsingValidation(unittest.TestCase):
    """ Test parsing text using the standalone parser, for valid and invalid
    text, without further evaluation on the meaning of their content.
    """

    def test_invalid_examples(self):
        bad_examples = [
            "",
            "1name",
            "a.b.c^abc",
            "[a.b]c",
            "a*.c",
            "a:[b,c]:",
            ".a",
            "a()",
            "-a",
        ]
        for bad_example in bad_examples:
            with self.subTest(bad_example=bad_example):
                with self.assertRaises(UnexpectedInput):
                    PARSER.parse(bad_example)

    def test_valid_examples(self):
        good_examples = [
            "name",
            "name123",
            "name_a",
            "_name",
            "foo.bar",
            "foo  .  bar",
            "foo:bar",
            "foo  :  bar",
            "foo,bar",
            "foo  ,  bar",
            "[foo,bar,foo.spam]",
            "[foo, bar].baz",
            "[foo, [bar, baz]]:spam",
            "foo:[bar.spam,baz]",
            "foo.items",
            "items",
            "+metadata_name",
        ]

        for good_example in good_examples:
            with self.subTest(good_example=good_example):
                try:
                    PARSER.parse(good_example)
                except Exception:
                    self.fail(
                        "Parsing {!r} is expected to succeed.".format(
                            good_example
                        )
                    )



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Unit test case for testing trait types created by subclassing TraitType.
"""

import os
import sys
import tempfile
import textwrap
import shutil
import subprocess
import unittest

from traits.api import (
    DefaultValue,
    Float,
    Function,
    Method,
    NoDefaultSpecified,
    Symbol,
    TraitType,
    Undefined
)
from traits.testing.optional_dependencies import requires_numpy


class TraitTypesTest(unittest.TestCase):
    def test_traits_shared_transient(self):
        # Regression test for a bug in traits where the same _metadata
        # dictionary was shared between different trait types.
        class LazyProperty(TraitType):

            #: The default value type to use.
            default_value_type = DefaultValue.constant

            def get(self, obj, name):
                return 1729

        self.assertFalse(Float().transient)
        LazyProperty().as_ctrait()
        self.assertFalse(Float().transient)

    @requires_numpy
    def test_numpy_validators_loaded_if_numpy_present(self):
        # If 'numpy' is available, the numpy validators should be loaded,
        # even if numpy is imported after traits.
        test_script = textwrap.dedent("""
            from traits.trait_types import bool_fast_validate
            import numpy

            if numpy.bool_ in bool_fast_validate:
                print("Success")
            else:
                print("Failure")
        """)
        this_python = sys.executable
        tmpdir = tempfile.mkdtemp()
        try:
            tmpfile = os.path.join(tmpdir, "test_script.py")
            with open(tmpfile, "w", encoding="utf-8") as f:
                f.write(test_script)
            cmd = [this_python, tmpfile]
            output = subprocess.check_output(cmd).decode("utf-8")
        finally:
            shutil.rmtree(tmpdir)

        self.assertEqual(output.strip(), "Success")

    def test_default_value_in_init(self):

        class MyTraitType(TraitType):
            pass

        # Tests for the behaviour of the default_value argument
        # to TraitType.__init__.
        trait_type = MyTraitType(default_value=23)
        self.assertEqual(
            trait_type.get_default_value(),
            (DefaultValue.constant, 23),
        )

        # An explicit default value of None should work as expected.
        trait_type = MyTraitType(default_value=None)
        self.assertEqual(
            trait_type.get_default_value(),
            (DefaultValue.constant, None),
        )

        # If no default is given, get_default_value() returns a value
        # of Undefined.
        trait_type = MyTraitType()
        self.assertEqual(
            trait_type.get_default_value(),
            (DefaultValue.constant, Undefined),
        )

        # Similarly, if NoDefaultSpecified is given, get_default_value()
        # is Undefined.
        trait_type = MyTraitType(default_value=NoDefaultSpecified)
        self.assertEqual(
            trait_type.get_default_value(),
            (DefaultValue.constant, Undefined),
        )

    def test_disallowed_default_value(self):
        class MyTraitType(TraitType):

            default_value_type = DefaultValue.disallow

        trait_type = MyTraitType()
        self.assertEqual(
            trait_type.get_default_value(),
            (DefaultValue.disallow, Undefined)
        )

        ctrait = trait_type.as_ctrait()
        self.assertEqual(
            ctrait.default_value(),
            (DefaultValue.disallow, Undefined),
        )
        self.assertEqual(ctrait.default_kind, "invalid")
        self.assertEqual(ctrait.default, Undefined)

        with self.assertRaises(ValueError):
            ctrait.default_value_for(None, "<dummy>")

    def test_call_sets_default_value_type(self):
        class FooTrait(TraitType):
            default_value_type = DefaultValue.callable_and_args

            def __init__(self, default_value=NoDefaultSpecified, **metadata):
                default_value = (pow, (3, 4), {})
                super().__init__(default_value, **metadata)

        trait = FooTrait()
        ctrait = trait.as_ctrait()
        self.assertEqual(ctrait.default_value_for(None, "dummy"), 81)
        cloned_ctrait = trait(30)
        self.assertEqual(cloned_ctrait.default_value_for(None, "dummy"), 30)


class TestDeprecatedTraitTypes(unittest.TestCase):
    def test_function_deprecated(self):
        def some_function():
            pass

        with self.assertWarnsRegex(DeprecationWarning, "Function trait type"):
            Function()
        with self.assertWarnsRegex(DeprecationWarning, "Function trait type"):
            Function(some_function, washable=True)

    def test_method_deprecated(self):

        class A:
            def some_method(self):
                pass

        with self.assertWarnsRegex(DeprecationWarning, "Method trait type"):
            Method()
        with self.assertWarnsRegex(DeprecationWarning, "Method trait type"):
            Method(A().some_method, gluten_free=False)

    def test_symbol_deprecated(self):
        with self.assertWarnsRegex(DeprecationWarning, "Symbol trait type"):
            Symbol("random:random")



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# --(Creating New Trait Types)-------------------------------------------------
"""
Creating New Trait Types
========================

You create a *new style* trait type by subclassing the **TraitType** class or
one of its subclasses, such as **Float** or **Str**. **TraitType** provides
the infrastructure for creating a trait type and allows subclasses to define
specific methods and class constants used to create either a new trait *type*
or *property*.

In the next section, we'll cover the methods and class constants used to define
a new trait *type*, and in the section following that we'll show how to define
a new trait *property*.

Defining a New Trait Type
-------------------------

The thing that distinguishes a trait *type* from a *property* is the existence
of a *validate* method in the subclass. The *validate* method should have the
following signature:

validate(self, object, name, value)
    This method validates, coerces, or adapts the specified *value* as the
    value of the *name* trait of the *object* object. This method is called
    when a value is assigned to an object trait that is based on this subclass
    of *TraitType* and the class does not  contain a definition for either the
    *get()* or *set()* methods.

    The method must return the original *value* or any suitably coerced or
    adapted value that is a legal value for the trait. If *value* is  not a
    legal value for the trait, and cannot be coerced or adapted to a legal
    value, the method should either raise a **TraitError** or  call the
    *error()* method to raise a **TraitError** on its behalf.

In addition to *validate*, the subclass can also define the *post_setattr*
method, which should have the following signature:

post_setattr(self, object, name, value)
    This method allows the trait to do additional processing after *value* has
    been successfully assigned to the *name* trait of the *object* object. For
    most traits there is no additional processing that needs to be done, and
    this method need not be defined. It is normally used for creating *shadow*
    (i.e., *mapped* traits), but other uses may arise as well. This method does
    not need to return a value, and should normally not raise any exceptions.

The subclass can also define a constant default value by setting the class-
level *default_value* attribute to the desired constant value. For example::

    class OddInt(BaseInt):

        default_value = 1

        ...

If a non-constant default value is desired, you should override the
**TraitType** class's *get_default_value* method. Refer to the documentation
for the **TraitType** class for more information on what this method does.

If you have a constant string that can be used as the type's *info* value, you
can provide it by simple setting the string as the value of the class-level
*info_text* attribute::

    class OddInt(BaseInt):

        info_text = 'an odd integer'

        ...

If you have a type info string which depends upon the internal state of the
trait, then you should override the **TraitType's** *info* method. This method
has no arguments, and should return a string describing the values accepted by
the trait type (e.g. 'an integer in the range from 1 to 5').

If you also have some type specific initialization that needs to be performed
when the trait type is created, you can also override the **TraitType** class's
*init* method. This method has no arguments and is automatically called from
the **TraitType** class constructor.

If you would like to specify a default Traits UI editor for your new trait
type, you can also override the **TraitType** class's *create_editor* method,
which has no arguments and should return the default **TraitEditor** for any
instances of the type to use.

This provides a basic overview of the basic methods and class constants needed
to define a new trait type. Refer to the complete documentation for the
**TraitType** and **BaseTraitHandler** classes for more information on other
methods that can be overridden if necessary.

Defining a New Trait Property
-----------------------------

You can also define new trait *properties* by subclassing from **TraitType** or
one of its subclasses. A *property* is distinguished from a *type* by the
existence of a *get* and/or *set* method in the **TraitType** subclass. The
signature for these two methods is as follows:

get(self, object, name)
    This is the *getter* method of a trait that behaves like a property. It has
    the following arguments:

    object
        The object that the property applies to.
    name
        The name of the *object* property.

    If this method is not defined, but the *set* method is defined, the trait
    behaves like a *write-only* property. This method should return the value
    of the *name* property for the *object* object.

set(self, object, name, value)
    This is the *setter* method of a trait that behaves like a property. It has
    the following arguments:

    object
        The object the property applies to.
    name
        The name of the property on *object*.
    value
        The value being assigned as the value of the property.

    If this method is not defined, but the *get* method is defined, the trait
    behaves like a *read-only* property. This method does not need to return a
    value, but it should raise a **TraitError** exception if the specified
    *value* is not valid and cannot be  coerced or adapted to a valid value.

Because the value of a *property* is determined by the *get* method, the
*default_value* class constant and *get_default_value* method are not used.

However, all other values and methods, such as the *info_text* class attribute
and *info* method apply to a *property* as well as a normal type. Please refer
to the preceding section on defining a trait type for additional information
that applies to properties as well.
"""

# --<Imports>------------------------------------------------------------------
from random import randint

from traits.api import HasTraits, List, TraitError, TraitType


# --[DiceRoll Type]------------------------------------------------------------
# Define a type whose value represents the roll of a pair of dice:
class DiceRoll(TraitType):

    # Set default value to 'snake-eyes':
    default_value = (1, 1)

    # Describe the type:
    info_text = (
        "a tuple of the form (n,m), where both n and m are integers "
        "in the range from 1 to 6 representing a roll of a pair of "
        "dice"
    )

    # Validate any value assigned to the trait to make sure it is a valid
    # dice roll:
    def validate(self, object, name, value):
        if (
            isinstance(value, tuple)
            and (len(value) == 2)
            and (1 <= value[0] <= 6)
            and (1 <= value[1] <= 6)
        ):
            return value

        self.error(object, name, value)


# --[RandInt Property]---------------------------------------------------------

# Define a read-only property whose value is a random integer in a specified
# range:
class RandInt(TraitType):

    # Define the type's constructor:
    def __init__(self, low=1, high=10, **metadata):
        super().__init__(**metadata)
        self.low = int(low)
        self.high = int(high)

    # Define the property's getter:
    def get(self):
        return randint(self.low, self.high)

    # Define the type's type information:
    def info(self):
        return "a random integer in the range from %d to %d" % (
            self.low,
            self.high,
        )


# --[Craps Class]--------------------------------------------------------------


# Define a test class containing both new trait types/properties:
class Craps(HasTraits):

    rolls = List(DiceRoll)
    die = RandInt(1, 6)


# --[Example*]-----------------------------------------------------------------

# Create a test object:
craps = Craps()

# Add a number of test dice rolls:
for i in range(10):
    craps.rolls.append((craps.die, craps.die))

# Display the results:
print(craps.rolls)

# Try to assign an invalid dice roll:
try:
    craps.rolls.append((0, 0))
except TraitError:
    print("Assigning an invalid dice roll failed.")



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# interface_implementation.py - Example of implementing an interface

# --[Imports]------------------------------------------------------------------
from traits.api import HasTraits, Instance, provides, Str
from interface_definition import IName


# --[Code]---------------------------------------------------------------------
@provides(IName)
class Person(HasTraits):
    first_name = Str("John")
    last_name = Str("Doe")

    # Implementation of the 'IName' interface:
    def get_name(self):
        """ Returns the name of an object. """
        return "%s %s" % (self.first_name, self.last_name)


# --[Example*]-----------------------------------------------------------------
class Apartment(HasTraits):
    renter = Instance(IName)


william = Person(first_name="William", last_name="Adams")
apt1 = Apartment(renter=william)
print("Renter is: ", apt1.renter.get_name())
# Result: Renter is: William Adams



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# metadata.py --- Example of accessing trait metadata attributes

# --[Imports]------------------------------------------------------------------
from traits.api import (
    HasTraits, Int, List, Float, Instance, Any, Str, TraitType)


# --[Code]---------------------------------------------------------------------
class Foo(HasTraits):
    pass


class Test(HasTraits):
    i = Int(99)
    lf = List(Float)
    foo = Instance(Foo, ())
    any = Any("123")


# --[Example*]-----------------------------------------------------------------
t = Test()

# Prints values of various metadata attributes for each of the traits.
print(t.trait("i").default)  # 99
print(t.trait("i").default_kind)  # value
print(t.trait("i").inner_traits)  # ()
print(t.trait("i").is_trait_type(Int))  # True
print(t.trait("i").is_trait_type(Float))  # False

print(t.trait("lf").default)  # []
print(t.trait("lf").default_kind)  # list
print(t.trait("lf").inner_traits)  # (<traits.traits.CTrait object at
#  0x01B24138>,)
print(t.trait("lf").is_trait_type(List))  # True
print(t.trait("lf").is_trait_type(TraitType))  # True
print(t.trait("lf").is_trait_type(Float))  # False
print(t.trait("lf").inner_traits[0].is_trait_type(Float))  # True

print(t.trait("foo").default)  # <undefined>
print(t.trait("foo").default_kind)  # factory
print(t.trait("foo").inner_traits)  # ()
print(t.trait("foo").is_trait_type(Instance))  # True
print(t.trait("foo").is_trait_type(List))  # False

print(t.trait("any").default)  # 123
print(t.trait("any").default_kind)  # value
print(t.trait("any").inner_traits)  # ()
print(t.trait("any").is_trait_type(Any))  # True
print(t.trait("any").is_trait_type(Str))  # False



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# keywords.py --- Example of trait keywords

# --[Imports]------------------------------------------------------------------
from traits.api import HasTraits, Str


# --[Code]---------------------------------------------------------------------
class Person(HasTraits):
    # 'label' is used for Traits UI field labels;
    # 'desc' can be used for tooltips.
    first_name = Str("", desc="first or personal name", label="First Name")
    last_name = Str("", desc="last or family name", label="Last Name")



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A function to import symbols. """

from importlib import import_module

from traits.trait_base import xgetattr


def import_symbol(symbol_path):
    """ Import the symbol defined by the specified symbol path.

    Examples
    --------

    import_symbol('tarfile:TarFile') -> TarFile
    import_symbol('tarfile:TarFile.open') -> TarFile.open

    To allow compatibility with old-school traits symbol names we also allow
    all-dotted paths, but in this case you can only import top-level names
    from the module.

    import_symbol('tarfile.TarFile') -> TarFile

    """

    if ":" in symbol_path:
        module_name, symbol_name = symbol_path.split(":")

        module = import_module(module_name)
        symbol = xgetattr(module, symbol_name)

    else:
        components = symbol_path.split(".")
        module_name = ".".join(components[:-1])
        symbol_name = components[-1]

        module = import_module(module_name)
        symbol = getattr(module, symbol_name)

    return symbol



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
    A Trait Documenter
    (Subclassed from the autodoc ClassLevelDocumenter)

"""
from importlib import import_module
import inspect
import io
import token
import tokenize
import traceback

from sphinx.ext.autodoc import ClassLevelDocumenter
from sphinx.util import logging

from traits.has_traits import MetaHasTraits
from traits.trait_type import TraitType
from traits.traits import generic_trait


logger = logging.getLogger(__name__)


def _is_class_trait(name, cls):
    """ Check if the name is in the list of class defined traits of ``cls``.
    """
    return (
        isinstance(cls, MetaHasTraits)
        and name in cls.__class_traits__
        and cls.__class_traits__[name] is not generic_trait
    )


class TraitDocumenter(ClassLevelDocumenter):
    """ Specialized Documenter subclass for trait attributes.

    The class defines a new documenter that recovers the trait definition
    signature of module level and class level traits.

    To use the documenter, append the module path in the extension
    attribute of the `conf.py`.
    """

    # ClassLevelDocumenter interface #####################################

    objtype = "traitattribute"
    directivetype = "attribute"
    member_order = 60

    # must be higher than other attribute documenters
    priority = 12

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        """ Check that the documented member is a trait instance.
        """
        check = (
            isattr
            and issubclass(type(member), TraitType)
            or _is_class_trait(membername, parent.object)
        )
        return check

    def document_members(self, all_members=False):
        """ Trait attributes have no members """
        pass

    def import_object(self):
        """ Get the Trait object.

        Notes
        -----
        Code adapted from autodoc.Documenter.import_object.

        """
        try:
            current = self.module = import_module(self.modname)
            for part in self.objpath[:-1]:
                current = self.get_attr(current, part)
            name = self.objpath[-1]
            self.object_name = name
            self.object = None
            self.parent = current
            return True
        # this used to only catch SyntaxError, ImportError and
        # AttributeError, but importing modules with side effects can raise
        # all kinds of errors.
        except Exception as err:
            if self.env.app and not self.env.app.quiet:
                self.env.app.info(traceback.format_exc().rstrip())
            msg = (
                "autodoc can't import/find {0} {r1}, it reported error: "
                '"{2}", please check your spelling and sys.path'
            )
            self.directive.warn(
                msg.format(self.objtype, str(self.fullname), err)
            )
            self.env.note_reread()
            return False

    def add_directive_header(self, sig):
        """ Add the directive header 'attribute' with the annotation
        option set to the trait definition.

        """
        ClassLevelDocumenter.add_directive_header(self, sig)
        try:
            definition = trait_definition(
                cls=self.parent,
                trait_name=self.object_name,
            )
        except ValueError:
            # Without this, a failure to find the trait definition aborts
            # the whole documentation build.
            logger.warning(
                "No definition for the trait {!r} could be found in "
                "class {!r}.".format(self.object_name, self.parent),
                exc_info=True)
            return

        # Workaround for enthought/traits#493: if the definition is multiline,
        # throw away all lines after the first.
        if "\n" in definition:
            definition = definition.partition("\n")[0] + " …"

        self.add_line("   :annotation: = {0}".format(definition), "<autodoc>")


def trait_definition(*, cls, trait_name):
    """ Retrieve the portion of the source defining a Trait attribute.

    For example, given a class::

        class MyModel(HasStrictTraits)
            foo = List(Int, [1, 2, 3])

    ``trait_definition(cls=MyModel, trait_name="foo")`` returns
    ``"List(Int, [1, 2, 3])"``.

    Parameters
    ----------
    cls : MetaHasTraits
        Class being documented.
    trait_name : str
        Name of the trait being documented.

    Returns
    -------
    str
        The portion of the source containing the trait definition. For
        example, for a class trait defined as ``"my_trait = Float(3.5)"``,
        the returned string will contain ``"Float(3.5)"``.

    Raises
    ------
    ValueError
        If *trait_name* doesn't appear as a class-level variable in the
        source.
    """
    # Get the class source and tokenize it.
    source = inspect.getsource(cls)
    string_io = io.StringIO(source)
    tokens = tokenize.generate_tokens(string_io.readline)

    # find the trait definition start
    trait_found = False
    name_found = False
    while not trait_found:
        item = next(tokens, None)
        if item is None:
            break
        if name_found and item[:2] == (token.OP, "="):
            trait_found = True
            continue
        if item[:2] == (token.NAME, trait_name):
            name_found = True

    if not trait_found:
        raise ValueError(
            "No trait definition for {!r} found in {!r}".format(
                trait_name, cls)
        )

    # Retrieve the trait definition.
    definition_tokens = _get_definition_tokens(tokens)
    definition = tokenize.untokenize(definition_tokens).strip()
    return definition


def _get_definition_tokens(tokens):
    """ Given the tokens, extracts the definition tokens.

    Parameters
    ----------
    tokens : iterator
        An iterator producing tokens.

    Returns
    -------
    A list of tokens for the definition.
    """
    # Retrieve the trait definition.
    definition_tokens = []
    first_line = None

    for type, name, start, stop, line_text in tokens:
        if first_line is None:
            first_line = start[0]

        if type == token.NEWLINE:
            break

        item = (
            type,
            name,
            (start[0] - first_line + 1, start[1]),
            (stop[0] - first_line + 1, stop[1]),
            line_text,
        )

        definition_tokens.append(item)

    return definition_tokens


def setup(app):
    """ Add the TraitDocumenter in the current sphinx autodoc instance. """
    app.add_autodocumenter(TraitDocumenter)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# prototype_prefix.py --- Examples of PrototypedFrom() prefix parameter

# --[Imports]------------------------------------------------------------------
from traits.api import PrototypedFrom, Float, HasTraits, Instance, Str


# --[Code]---------------------------------------------------------------------
class Parent(HasTraits):
    first_name = Str
    family_name = ""
    favorite_first_name = Str
    child_allowance = Float(1.00)


class Child(HasTraits):
    __prefix__ = "child_"
    first_name = PrototypedFrom("mother", "favorite_*")
    last_name = PrototypedFrom("father", "family_name")
    allowance = PrototypedFrom("father", "*")
    father = Instance(Parent)
    mother = Instance(Parent)


# --[Example*]-----------------------------------------------------------------
fred = Parent(
    first_name="Fred",
    family_name="Lopez",
    favorite_first_name="Diego",
    child_allowance=5.0,
)

maria = Parent(
    first_name="Maria",
    family_name="Gonzalez",
    favorite_first_name="Tomas",
    child_allowance=10.0,
)

nino = Child(father=fred, mother=maria)

print(
    "%s %s gets $%.2f for allowance"
    % (nino.first_name, nino.last_name, nino.allowance)
)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# object_trait_attrs.py --- Example of per-object trait attributes

# --[Imports]------------------------------------------------------------------
from traits.api import HasTraits, Range


# --[Code]---------------------------------------------------------------------
class GUISlider(HasTraits):
    def __init__(
        self,
        eval=None,
        label="Value",
        trait=None,
        min=0.0,
        max=1.0,
        initial=None,
        **traits
    ):
        HasTraits.__init__(self, **traits)
        if trait is None:
            if min > max:
                min, max = max, min
            if initial is None:
                initial = min
            elif not (min <= initial <= max):
                initial = [min, max][abs(initial - min) > abs(initial - max)]
            trait = Range(min, max, value=initial)
        self.add_trait(label, trait)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest
from unittest import mock

from traits.api import HasTraits, Int
from traits.observation._filtered_trait_observer import FilteredTraitObserver
from traits.observation._metadata_filter import MetadataFilter
from traits.observation._testing import (
    call_add_or_remove_notifiers,
    create_graph,
)


class TestMetadataFilter(unittest.TestCase):
    """ Test the behaviour of the MetadataFilter """

    def test_metadata_defined_vs_undefined(self):
        metadata_filter = MetadataFilter(
            metadata_name="name",
        )
        self.assertTrue(
            metadata_filter("name", Int(name=True).as_ctrait()),
            "Expected the filter to return true"
        )
        self.assertFalse(
            metadata_filter("name", Int().as_ctrait()),
            "Expected the filter to return false"
        )

    def test_metadata_defined_as_none_is_same_as_undefined(self):
        # If a metadata has a None value, it is equivalent to the metadata
        # not being defined.
        metadata_filter = MetadataFilter(
            metadata_name="name",
        )
        self.assertFalse(
            metadata_filter("name", Int(name=None).as_ctrait()),
            "Expected the filter to return false"
        )

    def test_filter_equality(self):
        filter1 = MetadataFilter(
            metadata_name="name",
        )
        filter2 = MetadataFilter(
            metadata_name="name",
        )
        self.assertEqual(filter1, filter2)
        self.assertEqual(hash(filter1), hash(filter2))

    def test_filter_not_equal_name_different(self):
        filter1 = MetadataFilter(
            metadata_name="number",
        )
        filter2 = MetadataFilter(
            metadata_name="name",
        )
        self.assertNotEqual(filter1, filter2)

    def test_filter_not_equal_different_type(self):
        filter1 = MetadataFilter(
            metadata_name="name",
        )
        imposter = mock.Mock()
        imposter.metadata_name = "name"
        self.assertNotEqual(imposter, filter1)

    def test_slots(self):
        filter = MetadataFilter(metadata_name="name")
        with self.assertRaises(AttributeError):
            filter.__dict__
        with self.assertRaises(AttributeError):
            filter.__weakref__

    def test_repr_value(self):
        metadata_filter = MetadataFilter(
            metadata_name="name",
        )
        actual = repr(metadata_filter)
        self.assertEqual(
            actual,
            "MetadataFilter(metadata_name='name')"
        )

    def test_eval_repr_roundtrip(self):
        metadata_filter = MetadataFilter(
            metadata_name="name",
        )
        self.assertEqual(eval(repr(metadata_filter)), metadata_filter)


class TestWithFilteredTraitObserver(unittest.TestCase):
    """ Test MetadataFilter with FilteredTraitObserver and HasTraits. """

    def test_filter_metadata(self):

        class Person(HasTraits):
            n_jobs = Int(status="public")
            n_children = Int()   # no metadata

        observer = FilteredTraitObserver(
            filter=MetadataFilter(
                metadata_name="status",
            ),
            notify=True,
        )
        person = Person()
        handler = mock.Mock()
        call_add_or_remove_notifiers(
            object=person,
            graph=create_graph(observer),
            handler=handler,
        )

        # when
        person.n_jobs += 1

        # then
        self.assertEqual(handler.call_count, 1)
        handler.reset_mock()

        # when
        person.n_children += 1

        # then
        self.assertEqual(handler.call_count, 0)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import os
import pickle
import pickletools
import shutil
import tempfile
import unittest
import unittest.mock as mock
import warnings

from traits.api import HasTraits, Int
from traits.testing.optional_dependencies import requires_traitsui, traitsui


class Model(HasTraits):
    count = Int


@requires_traitsui
class TestConfigureTraits(unittest.TestCase):
    def setUp(self):
        self.toolkit = traitsui.api.toolkit()
        self.tmpdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        del self.tmpdir
        del self.toolkit

    def test_simple_call(self):
        # Minimal exercising of configure_traits functionality.
        model = Model()
        with mock.patch.object(self.toolkit, "view_application") as mock_view:
            model.configure_traits()
        self.assertEqual(mock_view.call_count, 1)

    def test_filename_but_no_file(self):
        model = Model(count=37)
        filename = os.path.join(self.tmpdir, "nonexistent.pkl")
        self.assertFalse(os.path.exists(filename))

        with mock.patch.object(self.toolkit, "view_application"):
            with self.assertWarns(DeprecationWarning):
                model.configure_traits(filename=filename)

        self.assertTrue(os.path.exists(filename))
        with open(filename, "rb") as pickled_object:
            unpickled = pickle.load(pickled_object)

        self.assertIsInstance(unpickled, Model)
        self.assertEqual(unpickled.count, model.count)

    def test_pickle_protocol(self):
        # Pickled files should use protocol 3 (a compromise between
        # efficiency and wide applicability).

        model = Model(count=37)
        filename = os.path.join(self.tmpdir, "nonexistent.pkl")
        self.assertFalse(os.path.exists(filename))

        with mock.patch.object(self.toolkit, "view_application"):
            with self.assertWarns(DeprecationWarning):
                model.configure_traits(filename=filename)

        self.assertTrue(os.path.exists(filename))
        with open(filename, "rb") as pickled_object_file:
            pickled_object = pickled_object_file.read()

        # Get and check the first opcode
        opcode, arg, _ = next(pickletools.genops(pickled_object))
        self.assertEqual(opcode.name, "PROTO")
        self.assertEqual(arg, 3)

    def test_filename_with_existing_file(self):
        # Create pre-existing pickle file.
        stored_model = Model(count=52)
        filename = os.path.join(self.tmpdir, "model.pkl")
        with open(filename, "wb") as pickled_object:
            pickle.dump(stored_model, pickled_object)

        model = Model(count=19)
        with mock.patch.object(self.toolkit, "view_application"):
            with self.assertWarns(DeprecationWarning):
                model.configure_traits(filename=filename)
        self.assertEqual(model.count, 52)

    def test_filename_with_invalid_existing_file(self):
        # Create file whose contents are not unpickleable.
        filename = os.path.join(self.tmpdir, "model.pkl")
        with open(filename, "wb") as pickled_object:
            pickled_object.write(b"this is not a valid pickle")

        model = Model(count=19)
        with mock.patch.object(self.toolkit, "view_application"):
            with self.assertRaises(pickle.PickleError):
                with self.assertWarns(DeprecationWarning):
                    model.configure_traits(filename=filename)

    def test_filename_with_existing_file_stores_updated_model(self):
        stored_model = Model(count=52)
        filename = os.path.join(self.tmpdir, "model.pkl")
        with open(filename, "wb") as pickled_object:
            pickle.dump(stored_model, pickled_object)

        def modify_model(*args, **kwargs):
            model.count = 23
            return mock.DEFAULT

        model = Model(count=19)
        with mock.patch.object(self.toolkit, "view_application") as mock_view:
            mock_view.side_effect = modify_model
            with self.assertWarns(DeprecationWarning):
                model.configure_traits(filename=filename)
        self.assertEqual(model.count, 23)

        with open(filename, "rb") as pickled_object:
            unpickled = pickle.load(pickled_object)

        self.assertIsInstance(unpickled, Model)
        self.assertEqual(unpickled.count, model.count)

    def test_edit_when_false(self):
        # Check for deprecation warning when *edit* is false.
        model = Model()
        with mock.patch.object(self.toolkit, "view_application") as mock_view:
            mock_view.return_value = True
            with self.assertWarns(DeprecationWarning):
                model.configure_traits(edit=False)
        mock_view.assert_not_called()

    def test_edit_when_true(self):
        # Check for deprecation warning when *edit* is false.
        model = Model()
        with mock.patch.object(self.toolkit, "view_application") as mock_view:
            mock_view.return_value = True
            with self.assertWarns(DeprecationWarning):
                model.configure_traits(edit=True)
        mock_view.assert_called_once()

    def test_edit_not_given(self):
        # Normal case where the *edit* argument is not supplied.
        model = Model()
        with mock.patch.object(self.toolkit, "view_application") as mock_view:
            mock_view.return_value = True
            with warnings.catch_warnings(record=True) as captured_warnings:
                warnings.simplefilter("always", DeprecationWarning)
                model.configure_traits()
        mock_view.assert_called_once()

        all_warnings = "".join(
            str(warning.message) for warning in captured_warnings
        )
        self.assertNotIn("edit argument", all_warnings)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Manages all registered adaptations. """


from heapq import heappop, heappush
import inspect
import itertools
import functools

from traits.adaptation.adaptation_error import AdaptationError
from traits.has_traits import HasTraits
from traits.trait_types import Dict, List, Str


def no_adapter_necessary(adaptee):
    """ An adapter factory used to register that a protocol provides another.

    See 'register_provides' for details.

    """

    return adaptee


class AdaptationManager(HasTraits):
    """ Manages all registered adaptations. """

    #### 'AdaptationManager' class protocol ###################################

    @staticmethod
    def mro_distance_to_protocol(from_type, to_protocol):
        """ Return the distance in the MRO from 'from_type' to 'to_protocol'.

        If `from_type` provides `to_protocol`, returns the distance between
        `from_type` and the super-most class in the MRO hierarchy providing
        `to_protocol` (that's where the protocol was provided in the first
        place).

        If `from_type` does not provide `to_protocol`, return None.

        """

        if not AdaptationManager.provides_protocol(from_type, to_protocol):
            return None

        # We walk up the MRO hierarchy until the point where the `to_protocol`
        # is *no longer* provided. When we reach that point we know that the
        # previous class in the MRO is the one that provided the protocol in
        # the first place (e.g., the first super-class implementing an
        # interface).
        supertypes = inspect.getmro(from_type)[1:]

        distance = 0
        for t in supertypes:
            if AdaptationManager.provides_protocol(t, to_protocol):
                distance += 1

            # We have reached the point in the MRO where the protocol is no
            # longer provided.
            else:
                break

        return distance

    @staticmethod
    def provides_protocol(type_, protocol):
        """ Does the given type provide (i.e implement) a given protocol?

        Parameters
        ----------
        type_
            Python 'type'.
        protocol
            Either a regular Python class or a traits Interface.

        Returns
        -------
        result : bool
            True if the object provides the protocol, otherwise False.

        """
        return issubclass(type_, protocol)

    #### 'AdaptationManager' protocol #########################################

    def adapt(self, adaptee, to_protocol, default=AdaptationError):
        """ Attempt to adapt an object to a given protocol.

        If *adaptee* already provides (i.e. implements) the given protocol
        then it is simply returned unchanged.

        Otherwise, we try to build a chain of adapters that adapt *adaptee* to
        *to_protocol*. If no such adaptation is possible then either an
        AdaptationError is raised, or *default* is returned.

        Parameters
        ----------
        adaptee : object
            The object that we want to adapt.
        to_protocol : type or interface
            The protocol that the want to adapt *adaptee* to.
        default : object, optional
            Object to return if no adaptation is possible. If no default is
            provided, and adaptation fails, an ``AdaptationError`` is raised.

        Returns
        -------
        adapted_object : to_protocol
            The original adaptee adapted to the target protocol.

        Raises
        ------
        AdaptationError
            If adaptation is not possible, and no default is given.
        """

        # If the object already provides the given protocol then it is
        # simply returned.
        if self.provides_protocol(type(adaptee), to_protocol):
            result = adaptee

        # Otherwise, try adapting the object.
        else:
            result = self._adapt(adaptee, to_protocol)

        if result is None:
            if default is AdaptationError:
                raise AdaptationError(
                    "Could not adapt %r to %r" % (adaptee, to_protocol)
                )
            else:
                result = default

        return result

    def register_offer(self, offer):
        """ Register an offer to adapt from one protocol to another. """

        offers = self._adaptation_offers.setdefault(
            offer.from_protocol_name, []
        )
        offers.append(offer)

    def register_factory(self, factory, from_protocol, to_protocol):
        """ Register an adapter factory.

        This is a simply a convenience method that creates and registers an
        'AdaptationOffer' from the given arguments.

        """

        from traits.adaptation.adaptation_offer import AdaptationOffer

        self.register_offer(
            AdaptationOffer(
                factory=factory,
                from_protocol=from_protocol,
                to_protocol=to_protocol,
            )
        )

    def register_provides(self, provider_protocol, protocol):
        """ Register that a protocol provides another. """

        self.register_factory(
            no_adapter_necessary, provider_protocol, protocol
        )

    def supports_protocol(self, obj, protocol):
        """ Does the object support a given protocol?

        An object "supports" a protocol if either it "provides" it directly,
        or it can be adapted to it.

        """

        return self.adapt(obj, protocol, None) is not None

    #### Private protocol #####################################################

    #: All registered adaptation offers.
    #: Keys are the type name of the offer's from_protocol; values are a
    #: list of adaptation offers.
    _adaptation_offers = Dict(Str, List)

    def _adapt(self, adaptee, to_protocol):
        """ Returns an adapter that adapts an object to the target class.

        Returns None if no such adapter exists.

        """

        # The algorithm for finding a sequence of adapters adapting 'adaptee'
        # to 'to_protocol' is based on a weighted graph.

        # Nodes on the graphs are protocols (types or interfaces).
        # Edges are adaptation offers that connect a offer.from_protocol to a
        # offer.to_protocol.
        # Edges connect protocol A to protocol B and are weighted by two
        # numbers in this priority:
        # 1) a unit weight (1) representing the fact that we use 1 adaptation
        #    offer to go from A to B
        # 2) the number of steps up the type hierarchy that we need to take
        #    to go from A to offer.from_protocol, so that more specific
        #    adapters are always preferred

        # The algorithm finds the shortest weighted path between 'adaptee'
        # and 'to_protocol'. Once a candidate path is found, it tries to
        # create the adapters using the factories in the adaptation offers
        # that compose the path. If this fails because of conditional
        # adaptation (i.e., an adapter factory returns None), the path
        # is discarded and the algorithm looks for the next shortest path.

        # Cycles in adaptation are avoided by only considering path were
        # every adaptation offer is used at most once.

        # The implementation of the algorithm is based on a priority queue,
        # 'offer_queue'.
        #
        # Each value in the queue has got two parts,
        # one is the adaptation path, i.e., the sequence of adaptation offers
        # followed so far; the second value is the protocol of the last
        # visited node.
        #
        # The priority in the queue is the sum of all the weights for the
        # edges traversed in the path.

        # Unique sequence counter to make the priority list stable
        # w.r.t the sequence of insertion.
        counter = itertools.count()

        # The priority queue containing entries of the form
        # (cumulative weight, path, current protocol) describing an
        # adaptation path starting at `adaptee`, following a sequence
        # of adaptation offers, `path`, and having weight `cumulative_weight`.
        #
        # 'cumulative weight' is a tuple of the form
        # (number of traversed adapters,
        #  number of steps up protocol hierarchies,
        #  counter)
        #
        # The counter is an increasing number, and is used to make the
        # priority queue stable w.r.t insertion time
        # (see http://bit.ly/13VxILn).
        offer_queue = [((0, 0, next(counter)), [], type(adaptee))]

        while len(offer_queue) > 0:
            # Get the most specific candidate path for adaptation.
            weight, path, current_protocol = heappop(offer_queue)

            edges = self._get_applicable_offers(current_protocol, path)

            # Sort by weight first, then by from_protocol type.
            edges.sort(
                key=functools.cmp_to_key(
                    _by_weight_then_from_protocol_specificity
                )
            )

            # At this point, the first edges are the shortest ones. Within
            # edges with the same distance, interfaces which are subclasses
            # of other interfaces in that group come first. The rest of
            # the order is unspecified.

            for mro_distance, offer in edges:
                new_path = path + [offer]

                # Check if we arrived at the target protocol.
                if self.provides_protocol(offer.to_protocol, to_protocol):
                    # Walk path and create adapters
                    adapter = adaptee
                    for offer in new_path:
                        adapter = offer.factory(adapter)
                        if adapter is None:
                            # This adaptation attempt failed (e.g. because of
                            # conditional adaptation).
                            # Discard this path and continue.
                            break

                    else:
                        # We're done!
                        return adapter

                else:
                    # Push the new path on the priority queue.
                    adapter_weight, mro_weight, _ = weight
                    new_weight = (
                        adapter_weight + 1,
                        mro_weight + mro_distance,
                        next(counter),
                    )
                    heappush(
                        offer_queue, (new_weight, new_path, offer.to_protocol)
                    )

        return None

    def _get_applicable_offers(self, current_protocol, path):
        """ Find all adaptation offers that can be applied to a protocol.

        Return all the applicable offers together with the number of steps
        up the MRO hierarchy that need to be taken from the protocol
        to the offer's from_protocol.
        The returned object is a list of tuples (mro_distance, offer) .

        In terms of our graph algorithm, we're looking for all outgoing edges
        from the current node.
        """

        edges = []

        for from_protocol_name, offers in self._adaptation_offers.items():
            from_protocol = offers[0].from_protocol
            mro_distance = self.mro_distance_to_protocol(
                current_protocol, from_protocol
            )

            if mro_distance is not None:

                for offer in offers:
                    # Avoid cycles by checking that we did not consider this
                    # offer in this path.
                    if offer not in path:
                        edges.append((mro_distance, offer))

        return edges


def _by_weight_then_from_protocol_specificity(edge_1, edge_2):
    """ Comparison function for graph edges.

    Each edge is of the form (mro distance, adaptation offer).

    Comparison is done by mro distance first, and by offer's from_protocol
    issubclass next.

    If two edges have the same mro distance, and the from_protocols of the
    two edges are not subclasses of one another, they are considered "equal".

    """

    # edge_1 and edge_2 are edges, of the form (mro_distance, offer)

    mro_distance_1, offer_1 = edge_1
    mro_distance_2, offer_2 = edge_2

    # First, compare the MRO distance.
    if mro_distance_1 < mro_distance_2:
        return -1
    elif mro_distance_1 > mro_distance_2:
        return 1

    # The distance is equal, prefer more specific 'from_protocol's
    if offer_1.from_protocol is offer_2.from_protocol:
        return 0

    if issubclass(offer_1.from_protocol, offer_2.from_protocol):
        return -1
    elif issubclass(offer_2.from_protocol, offer_1.from_protocol):
        return 1

    return 0


#: The default global adaptation manager.
#:
#: PROVIDED FOR BACKWARD COMPATIBILITY ONLY, IT SHOULD NEVER BE USED DIRECTLY.
#: If you must use a global adaptation manager, use the functions
#: :class:`get_global_adaptation_manager`,
#: :class:`reset_global_adaptation_manager`,
#: :class:`set_global_adaptation_manager`.

adaptation_manager = AdaptationManager()


def set_global_adaptation_manager(new_adaptation_manager):
    """ Set the global adaptation manager to the given instance. """
    global adaptation_manager
    adaptation_manager = new_adaptation_manager


def reset_global_adaptation_manager():
    """ Set the global adaptation manager to a new AdaptationManager instance.
    """
    global adaptation_manager
    adaptation_manager = AdaptationManager()


def get_global_adaptation_manager():
    """ Set a reference to the global adaptation manager. """
    global adaptation_manager
    return adaptation_manager


# Convenience references to methods on the default adaptation manager.
#
# If you add a public method to the adaptation manager protocol then don't
# forget to add a convenience function here!


def adapt(adaptee, to_protocol, default=AdaptationError):
    """ Attempt to adapt an object to a given protocol. """
    manager = get_global_adaptation_manager()
    return manager.adapt(adaptee, to_protocol, default)


def register_factory(factory, from_protocol, to_protocol):
    """ Register an adapter factory. """
    manager = get_global_adaptation_manager()
    return manager.register_factory(factory, from_protocol, to_protocol)


def register_offer(offer):
    """ Register an offer to adapt from one protocol to another. """
    manager = get_global_adaptation_manager()
    return manager.register_offer(offer)


def register_provides(provider_protocol, protocol):
    """ Register that a protocol provides another. """
    manager = get_global_adaptation_manager()
    return manager.register_provides(provider_protocol, protocol)


def supports_protocol(obj, protocol):
    """ Does the object support a given protocol? """
    manager = get_global_adaptation_manager()
    return manager.supports_protocol(obj, protocol)


def provides_protocol(type_, protocol):
    """ Does the given type provide (i.e implement) a given protocol? """
    return AdaptationManager.provides_protocol(type_, protocol)



# (C) Copyright 2020-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


from collections import defaultdict
import os
import shutil
import re
import tempfile


def parse_py_file(filepath):
    """ This function parses a python file that have been annotated with error
    codes that mypy may generate and returns a mapping from line number to
    a list of those codes.

    File annotations may be done by adding a comment to the line that begins
    with "E:" followed by a list of comma separated error codes
    Eg::

        obj.var = False  # E: assignment, list-item

    Parameters
    ----------
    filepath : pathlib.Path
        The filepath of the python file

    Returns
    -------
    line_err_map : dict
        A mapping from line number to the list of error codes on that line
    """
    line_err_map = {}
    regex = re.compile(r'#[\s]*E:[\s]* (.*)')

    with filepath.open(encoding="utf-8") as fp:
        line_count = 0
        lines = fp.readlines()
        for line in lines:
            line_count += 1
            match = regex.search(line)
            if match:
                err_codes = match.group(1)
                list_err_codes = err_codes.replace(' ', '').split(',')
                line_err_map[line_count] = list_err_codes
    return line_err_map


def parse_mypy_output(output_str):
    """ Parses the output generated by mypy and returns a dict of mappings from
    line number to a list of error codes

    Parameters
    ----------
    output_str: str
        The output generated by mypy on stdout.

    Returns
    -------
    line_errors_dict : dict
        A mapping from line number to a list of error codes

    """
    line_errors_dict = defaultdict(set)
    error_line_regex = re.compile(r"([0-9]*): error:.*\[(.*)\]")
    for line in output_str.split("\n"):
        match = error_line_regex.search(line)
        if match:
            line_no_str, err_code = match.groups()
            line_errors_dict[int(line_no_str)].add(err_code)
    return line_errors_dict


def run_mypy(filepath):
    """ Runs mypy on the file using mypy api and returns the generated output.

    Parameters
    ----------
    filepath : pathlib.path
        The path to of the file to run mypy on.

    Returns
    -------
    normal_report : str
        Output on stdout
    error_report : str
        Output on stderr
    exit_status : int
        The exit status

    """
    # Local import to make it easier to skip tests if mypy is not in
    # the environment.
    from mypy import api as mypy_api

    # Need to use  tempdir since mypy complains that:
    # "site-packages is in PYTHONPATH. Please change directory so it is not."
    filepath = str(filepath)

    with tempfile.TemporaryDirectory() as tempdir:
        dest_filename = os.path.basename(filepath)
        dest = shutil.copyfile(filepath, os.path.join(tempdir, dest_filename))
        normal_report, error_report, exit_status = mypy_api.run(
            [dest, '--show-error-code'])
    return normal_report, error_report, exit_status


class MypyAssertions:

    def assertNoMypyError(self, filepath):
        """ Raises an AssertionError if mypy raises any error.

        Parameters
        ----------
        filepath : pathlib.Path
            The path to the file to run mypy on.

        Returns
        -------
        None

        Raises
        ------
        AssertionError
            If my generates any error

        """
        normal_report, error_report, exit_status = run_mypy(filepath)
        if exit_status != 0:
            s = "\n{}\n{}".format(str(filepath), normal_report)
            raise AssertionError(s)

    def assertRaisesMypyError(self, filepath):
        """ Raises an AssertionError if the errors raised by mypy are different
        from the expected errors annotated in the file. Any difference from
        expected errors, whether more or less errors will raise an
        Assertion Error

        Parameters
        ----------
        filepath : pathlib.Path
            The path to the file to run mypy on.

        Returns
        -------
        None

        Raises
        ------
        AssertionError
            If errors generated by mypy are different from expected errors.

        """
        line_error_map = parse_py_file(filepath)
        normal_report, error_report, exit_status = run_mypy(filepath)
        parsed_mypy_output = parse_mypy_output(normal_report)

        for line, error_codes in parsed_mypy_output.items():
            if line not in line_error_map or sorted(
                    line_error_map[line]) != sorted(list(error_codes)):
                s = "\n{}\n{}".format(str(filepath), normal_report)
                raise AssertionError(s)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.api import AbstractViewElement, HasTraits, Int, TraitError
from traits.testing.optional_dependencies import requires_traitsui


@requires_traitsui
class TestViewElements(unittest.TestCase):
    def setUp(self):
        import traitsui.api

        self.toolkit = traitsui.api.toolkit()

    def tearDown(self):
        del self.toolkit

    def test_view_definition(self):
        from traitsui.api import View

        view = View('count')

        class Model(HasTraits):
            count = Int
            my_view = view

        view_elements = Model.class_trait_view_elements()

        self.assertEqual(view_elements.content, {'my_view': view})

    def test_view_definition_twice(self):
        from traitsui.api import View

        view = View('count')

        class Model(HasTraits):
            count = Int
            my_view = view

        view_elements = Model.class_trait_view_elements()
        view_elements2 = Model.class_trait_view_elements()

        self.assertEqual(view_elements.content, {'my_view': view})
        self.assertIs(view_elements, view_elements2)

    def test_view_elements_parents(self):
        from traitsui.api import View

        class Model(HasTraits):
            count = Int
            my_view = View('count')

        class ModelSubclass(Model):
            total = Int
            my_view = View('count', 'total')

        view_elements = ModelSubclass.class_trait_view_elements()
        parent_view_elements = Model.class_trait_view_elements()

        self.assertEqual(view_elements.parents[0], parent_view_elements)

    def test_instance_view_definition(self):
        from traitsui.api import View

        view = View('count')

        class Model(HasTraits):
            count = Int
            my_view = view

        m = Model()
        view_elements = m.trait_view_elements()

        self.assertEqual(view_elements.content, {'my_view': view})

    def test_trait_views(self):
        from traitsui.api import View

        view = View('count')

        class Model(HasTraits):
            count = Int
            my_view = view

        m = Model()
        views = m.trait_views()

        self.assertEqual(views, ['my_view'])

    def test_included_names(self):
        from traitsui.api import Group, Item, View

        item = Item('count', id='item_with_id')
        group = Group(item)
        view = View(Item('count'))

        class Model(HasTraits):
            count = Int
            my_group = group
            my_view = view

        view_elements = Model.class_trait_view_elements()

        self.assertEqual(
            view_elements.content,
            {'my_view': view, 'my_group': group, 'item_with_id': item}
        )

    def test_duplicate_names(self):
        from traitsui.api import Group, Item, View

        class Model(HasTraits):
            count = Int
            includable = Group(Item('count', id='name_conflict'))
            name_conflict = View(Item('count'))

        with self.assertRaises(TraitError):
            Model.class_trait_view_elements()

    def test_view_element_superclass(self):
        from traitsui.api import ViewElement

        self.assertIsInstance(ViewElement(), AbstractViewElement)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.adaptation.api import reset_global_adaptation_manager
from traits.api import (
    BaseInstance,
    Bool,
    HasTraits,
    Instance,
    List,
    register_factory,
    TraitError,
)


class Foo(HasTraits):
    default = Bool(False)


class Bar(HasTraits):
    pass


def bar_to_foo_adapter(bar):
    return Foo()


def default_foo():
    return Foo(default=True)


class AutomaticAdaptationTestMixin:
    """
    Mixin for tests to be applied to both Instance and BaseInstance.

    Subclasses should define the class variable 'trait_under_test'.
    """

    #### 'TestCase' protocol ##################################################

    def setUp(self):
        reset_global_adaptation_manager()

    #### Tests ################################################################

    def test_instance_trait_automatic_adaptation(self):
        bar = Bar()
        foo_container = self.create_foo_container()

        # Before a Bar->Foo adapter is registered.
        with self.assertRaises(TraitError):
            foo_container.not_adapting_foo = bar

        with self.assertRaises(TraitError):
            foo_container.adapting_foo = bar

        foo_container.adapting_foo_permissive = bar
        self.assertIsNone(foo_container.adapting_foo_permissive)

        foo_container.adapting_foo_dynamic_default = bar
        self.assertIsInstance(foo_container.adapting_foo_dynamic_default, Foo)
        self.assertTrue(foo_container.adapting_foo_dynamic_default.default)

        # After a Bar->Foo adapter is registered.
        register_factory(bar_to_foo_adapter, Bar, Foo)

        with self.assertRaises(TraitError):
            foo_container.not_adapting_foo = bar

        foo_container.adapting_foo = bar
        self.assertIsInstance(foo_container.adapting_foo, Foo)

        foo_container.adapting_foo_permissive = bar
        self.assertIsInstance(foo_container.adapting_foo_permissive, Foo)

        foo_container.adapting_foo_dynamic_default = bar
        self.assertIsInstance(foo_container.adapting_foo_dynamic_default, Foo)
        self.assertFalse(foo_container.adapting_foo_dynamic_default.default)

    def test_list_trait_automatic_adaptation(self):
        bar = Bar()
        foo_container = self.create_foo_container()

        # Before a Bar->Foo adapter is registered.
        with self.assertRaises(TraitError):
            foo_container.not_adapting_foo_list = [bar]

        with self.assertRaises(TraitError):
            foo_container.adapting_foo_list = [bar]

        # After a Bar->Foo adapter is registered.
        register_factory(bar_to_foo_adapter, Bar, Foo)

        with self.assertRaises(TraitError):
            foo_container.not_adapting_foo_list = [bar]

        foo_container.adapting_foo_list = [bar]
        self.assertIsInstance(foo_container.adapting_foo_list[0], Foo)

    #### Helpers ##############################################################

    def create_foo_container(self):

        instance_trait = self.trait_under_test

        class FooContainer(HasTraits):
            not_adapting_foo = instance_trait(Foo)
            adapting_foo = instance_trait(Foo, adapt="yes")
            adapting_foo_permissive = instance_trait(Foo, adapt="default")
            adapting_foo_dynamic_default = instance_trait(
                Foo,
                adapt="default",
                factory=default_foo,
            )
            not_adapting_foo_list = List(Foo)
            adapting_foo_list = List(instance_trait(Foo, adapt="yes"))

        return FooContainer()


class TestAutomaticAdaptationInstance(
    AutomaticAdaptationTestMixin, unittest.TestCase
):
    """
    Tests for automatic adaptation with Instance.
    """

    #: Trait type being tested
    trait_under_test = Instance


class TestAutomaticAdaptationBaseInstance(
    AutomaticAdaptationTestMixin, unittest.TestCase
):
    """
    Tests for automatic adaptation with BaseInstance.
    """

    #: Trait type being tested
    trait_under_test = BaseInstance



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import inspect
import unittest

from traits.observation import expression
from traits.observation._anytrait_filter import anytrait_filter
from traits.observation._dict_item_observer import DictItemObserver
from traits.observation._filtered_trait_observer import FilteredTraitObserver
from traits.observation._list_item_observer import ListItemObserver
from traits.observation._metadata_filter import MetadataFilter
from traits.observation._named_trait_observer import NamedTraitObserver
from traits.observation._set_item_observer import SetItemObserver
from traits.observation._observer_graph import ObserverGraph


def create_graph(*nodes):
    """ Create an ObserverGraph with the given nodes joined one after another.
    Parameters
    ----------
    *nodes : hashable
        Items to be attached as nodes

    Returns
    -------
    ObserverGraph
    """
    node = nodes[-1]
    graph = ObserverGraph(node=node)
    for node in nodes[:-1][::-1]:
        graph = ObserverGraph(node=node, children=[graph])
    return graph


def create_expression(observer):
    """ Create an expression with a dummy observer for testing purposes.

    Parameters
    ----------
    observer : hashable
        Item to be used as a node on ObserverGraph

    Returns
    -------
    expression : ObserverExpression
    """
    return expression.SingleObserverExpression(observer)


class TestObserverExpressionComposition(unittest.TestCase):
    """ Test composition of ObserverExpression with generic observers."""

    def test_new_with_branches(self):
        observer = 1
        expr = create_expression(observer)
        expected = [
            create_graph(observer),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_or_operator(self):
        observer1 = 1
        observer2 = 2
        expr1 = create_expression(observer1)
        expr2 = create_expression(observer2)

        expr = expr1 | expr2
        expected = [
            create_graph(observer1),
            create_graph(observer2),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_or_maintain_order(self):
        # Test __or__ will maintain the order provided by the user.
        observer1 = 1
        observer2 = 2
        expr1 = create_expression(observer1)
        expr2 = create_expression(observer2)

        combined1 = expr1 | expr2
        combined2 = expr2 | expr1

        self.assertEqual(combined1._as_graphs(), combined2._as_graphs()[::-1])

    def test_then_operator(self):
        observer1 = 1
        observer2 = 2
        expr1 = create_expression(observer1)
        expr2 = create_expression(observer2)
        expr = expr1.then(expr2)

        expected = [
            create_graph(
                observer1,
                observer2,
            )
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_chained_then_or(self):
        observer1 = 1
        observer2 = 2
        observer3 = 3
        observer4 = 4
        expr1 = create_expression(observer1)
        expr2 = create_expression(observer2)
        expr3 = create_expression(observer3)
        expr4 = create_expression(observer4)

        expr = (expr1.then(expr2)) | (expr3.then(expr4))

        expected = [
            create_graph(
                observer1,
                observer2,
            ),
            create_graph(
                observer3,
                observer4,
            ),

        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_or_then_chained(self):
        observer1 = 1
        observer2 = 2
        observer3 = 3
        observer4 = 4
        expr1 = create_expression(observer1)
        expr2 = create_expression(observer2)
        expr3 = create_expression(observer3)
        expr4 = create_expression(observer4)

        expr = (expr1 | expr2).then(expr3 | expr4)

        expected = [
            ObserverGraph(
                node=observer1,
                children=[
                    create_graph(observer3),
                    create_graph(observer4),
                ],
            ),
            ObserverGraph(
                node=observer2,
                children=[
                    create_graph(observer3),
                    create_graph(observer4),
                ],
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_join_expressions(self):
        observer1 = 1
        observer2 = 2
        expr1 = create_expression(observer1)
        expr2 = create_expression(observer2)

        expr = expression.join(expr1, expr2)

        expected = [
            create_graph(
                observer1,
                observer2,
            )
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)


class TestObserverExpressionAnytrait(unittest.TestCase):
    """ Test anytrait function and method. """

    def test_anytrait_function_notify_true(self):
        expr = expression.anytrait(notify=True)
        expected = [
            create_graph(
                FilteredTraitObserver(filter=anytrait_filter, notify=True),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_anytrait_function_notify_false(self):
        expr = expression.anytrait(notify=False)
        expected = [
            create_graph(
                FilteredTraitObserver(filter=anytrait_filter, notify=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_anytrait_method_notify_true(self):
        expr = expression.trait("name").anytrait(notify=True)
        expected = [
            create_graph(
                NamedTraitObserver(name="name", notify=True, optional=False),
                FilteredTraitObserver(filter=anytrait_filter, notify=True),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_anytrait_method_notify_false(self):
        expr = expression.trait("name").anytrait(notify=False)
        expected = [
            create_graph(
                NamedTraitObserver(name="name", notify=True, optional=False),
                FilteredTraitObserver(filter=anytrait_filter, notify=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)


class TestObserverExpressionFilter(unittest.TestCase):
    """ Test ObserverExpression.match """

    def setUp(self):

        def anytrait(name, trait):
            return True

        self.anytrait = anytrait

    def test_match_notify_true(self):
        # Test the top-level function
        expr = expression.match(filter=self.anytrait)
        expected = [
            create_graph(
                FilteredTraitObserver(filter=self.anytrait, notify=True),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_match_notify_false(self):
        # Test the top-level function
        expr = expression.match(filter=self.anytrait, notify=False)
        expected = [
            create_graph(
                FilteredTraitObserver(filter=self.anytrait, notify=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_match_method_notify_true(self):
        # Test the instance method calls the top-level function correctly.
        expr = expression.match(filter=self.anytrait).match(
            filter=self.anytrait
        )
        expected = [
            create_graph(
                FilteredTraitObserver(filter=self.anytrait, notify=True),
                FilteredTraitObserver(filter=self.anytrait, notify=True),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_match_method_notify_false(self):
        # Test the instance method calls the top-level function correctly.
        expr = expression.match(filter=self.anytrait).match(
            filter=self.anytrait, notify=False,
        )
        expected = [
            create_graph(
                FilteredTraitObserver(filter=self.anytrait, notify=True),
                FilteredTraitObserver(filter=self.anytrait, notify=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_call_signatures(self):
        # Test to help developers keeping the two function signatures in-sync.
        # Remove this if the two need to divert in the future.
        top_level = expression.match
        method = expression.ObserverExpression().match
        self.assertEqual(
            inspect.signature(top_level), inspect.signature(method)
        )


class TestObserverExpressionFilterMetadata(unittest.TestCase):
    """ Test ObserverExpression.metadata """

    def test_metadata_notify_true(self):
        # Test the top-level function
        expr = expression.metadata("butterfly")
        expected = [
            create_graph(
                FilteredTraitObserver(
                    filter=MetadataFilter(metadata_name="butterfly"),
                    notify=True,
                ),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_metadata_notify_false(self):
        # Test the top-level function
        expr = expression.metadata("butterfly", notify=False)
        expected = [
            create_graph(
                FilteredTraitObserver(
                    filter=MetadataFilter(metadata_name="butterfly"),
                    notify=False,
                ),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_metadata_method_notify_true(self):
        # Test the instance method calls the top-level function correctly.
        expr = expression.metadata("bee").metadata("ant")
        expected = [
            create_graph(
                FilteredTraitObserver(
                    filter=MetadataFilter(metadata_name="bee"),
                    notify=True,
                ),
                FilteredTraitObserver(
                    filter=MetadataFilter(metadata_name="ant"),
                    notify=True,
                ),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_metadata_method_notify_false(self):
        # Test the instance method calls the top-level function correctly.
        expr = expression.metadata("bee").metadata("ant", notify=False)
        expected = [
            create_graph(
                FilteredTraitObserver(
                    filter=MetadataFilter(metadata_name="bee"),
                    notify=True,
                ),
                FilteredTraitObserver(
                    filter=MetadataFilter(metadata_name="ant"),
                    notify=False,
                ),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_call_signatures(self):
        # Test to help developers keeping the two function signatures in-sync.
        # Remove this if the two need to divert in the future.
        top_level = expression.metadata
        method = expression.ObserverExpression().metadata
        self.assertEqual(
            inspect.signature(top_level), inspect.signature(method)
        )


class TestObserverExpressionTrait(unittest.TestCase):
    """ Test ObserverExpression.trait """

    def test_trait_name(self):
        # Test the top-level function
        expr = expression.trait("name")
        expected = [
            create_graph(
                NamedTraitObserver(name="name", notify=True, optional=False)
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_trait_name_notify_false(self):
        # Test the top-level function
        expr = expression.trait("name", notify=False)
        expected = [
            create_graph(
                NamedTraitObserver(name="name", notify=False, optional=False)
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_trait_name_optional_true(self):
        # Test the top-level function
        expr = expression.trait("name", optional=True)
        expected = [
            create_graph(
                NamedTraitObserver(name="name", notify=True, optional=True)
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_trait_method(self):
        # Test the instance method calls the top-level function correctly.
        expr = expression.trait("name").trait("attr")
        expected = [
            create_graph(
                NamedTraitObserver(name="name", notify=True, optional=False),
                NamedTraitObserver(name="attr", notify=True, optional=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_trait_method_notify_false(self):
        # Test the instance method calls the top-level function correctly.
        expr = expression.trait("name").trait("attr", notify=False)
        expected = [
            create_graph(
                NamedTraitObserver(name="name", notify=True, optional=False),
                NamedTraitObserver(name="attr", notify=False, optional=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_trait_method_optional_true(self):
        # Test the instance method calls the top-level function correctly.
        expr = expression.trait("name").trait("attr", optional=True)
        expected = [
            create_graph(
                NamedTraitObserver(name="name", notify=True, optional=False),
                NamedTraitObserver(name="attr", notify=True, optional=True),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_call_signatures(self):
        # Test to help developers keeping the two function signatures in-sync.
        # Remove this if the two need to divert in the future.
        top_level_trait = expression.trait
        method_trait = expression.ObserverExpression().trait
        self.assertEqual(
            inspect.signature(top_level_trait), inspect.signature(method_trait)
        )


class TestObserverExpressionDictItem(unittest.TestCase):
    """ Test ObserverExpression.dict_items """

    def test_dict_items(self):
        expr = expression.dict_items()
        expected = [
            create_graph(
                DictItemObserver(notify=True, optional=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_dict_items_notify_false(self):
        expr = expression.dict_items(notify=False)
        expected = [
            create_graph(
                DictItemObserver(notify=False, optional=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_dict_items_optional_true(self):
        expr = expression.dict_items(optional=True)
        expected = [
            create_graph(
                DictItemObserver(notify=True, optional=True),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_dict_items_method_notify(self):
        # Test the instance method calls the top-level function correctly.
        expr = expression.dict_items().dict_items(notify=False)
        expected = [
            create_graph(
                DictItemObserver(notify=True, optional=False),
                DictItemObserver(notify=False, optional=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_dict_items_method_optional(self):
        # Test the instance method calls the top-level function correctly.
        expr = expression.dict_items().dict_items(optional=True)
        expected = [
            create_graph(
                DictItemObserver(notify=True, optional=False),
                DictItemObserver(notify=True, optional=True),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_call_signatures(self):
        # Test to help developers keeping the two function signatures in-sync.
        # Remove this if the two need to divert in the future.
        top_level = expression.dict_items
        method = expression.ObserverExpression().dict_items
        self.assertEqual(
            inspect.signature(top_level), inspect.signature(method)
        )


class TestObserverExpressionListItem(unittest.TestCase):
    """ Test ObserverExpression.list_items """

    def test_list_items(self):
        expr = expression.list_items()
        expected = [
            create_graph(
                ListItemObserver(notify=True, optional=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_list_items_notify_false(self):
        expr = expression.list_items(notify=False)
        expected = [
            create_graph(
                ListItemObserver(notify=False, optional=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_list_items_optional_true(self):
        expr = expression.list_items(optional=True)
        expected = [
            create_graph(
                ListItemObserver(notify=True, optional=True),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_list_items_method_notify(self):
        # Test the instance method calls the top-level function correctly.
        expr = expression.list_items().list_items(notify=False)
        expected = [
            create_graph(
                ListItemObserver(notify=True, optional=False),
                ListItemObserver(notify=False, optional=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_list_items_method_optional(self):
        # Test the instance method calls the top-level function correctly.
        expr = expression.list_items().list_items(optional=True)
        expected = [
            create_graph(
                ListItemObserver(notify=True, optional=False),
                ListItemObserver(notify=True, optional=True),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_call_signatures(self):
        # Test to help developers keeping the two function signatures in-sync.
        # Remove this if the two need to divert in the future.
        top_level = expression.list_items
        method = expression.ObserverExpression().list_items
        self.assertEqual(
            inspect.signature(top_level), inspect.signature(method)
        )


class TestObserverExpressionSetItem(unittest.TestCase):
    """ Test ObserverExpression.set_items """

    def test_set_items(self):
        expr = expression.set_items()
        expected = [
            create_graph(
                SetItemObserver(notify=True, optional=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_set_items_notify_false(self):
        expr = expression.set_items(notify=False)
        expected = [
            create_graph(
                SetItemObserver(notify=False, optional=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_set_items_optional_true(self):
        expr = expression.set_items(optional=True)
        expected = [
            create_graph(
                SetItemObserver(notify=True, optional=True),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_set_items_method_notify(self):
        # Test the instance method calls the top-level function correctly.
        expr = expression.set_items().set_items(notify=False)
        expected = [
            create_graph(
                SetItemObserver(notify=True, optional=False),
                SetItemObserver(notify=False, optional=False),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_set_items_method_optional(self):
        # Test the instance method calls the top-level function correctly.
        expr = expression.set_items().set_items(optional=True)
        expected = [
            create_graph(
                SetItemObserver(notify=True, optional=False),
                SetItemObserver(notify=True, optional=True),
            ),
        ]
        actual = expr._as_graphs()
        self.assertEqual(actual, expected)

    def test_call_signatures(self):
        # Test to help developers keeping the two function signatures in-sync.
        # Remove this if the two need to divert in the future.
        top_level = expression.set_items
        method = expression.ObserverExpression().set_items
        self.assertEqual(
            inspect.signature(top_level), inspect.signature(method)
        )


class TestObserverExpressionEqualityAndHashing(unittest.TestCase):
    """ Test ObserverExpression.__eq__ and ObserverExpression.__hash__. """

    def test_trait_equality(self):
        expr1 = create_expression(1)
        expr2 = create_expression(1)
        self.assertEqual(expr1, expr2)
        self.assertEqual(hash(expr1), hash(expr2))

    def test_join_equality_with_then(self):
        # The following all result in the same graphs
        expr1 = create_expression(1)
        expr2 = create_expression(2)

        combined1 = expression.join(expr1, expr2)
        combined2 = expr1.then(expr2)

        self.assertEqual(combined1, combined2)
        self.assertEqual(hash(combined1), hash(combined2))

    def test_equality_of_parallel_expressions(self):
        expr1 = create_expression(1) | create_expression(2)
        expr2 = create_expression(1) | create_expression(2)
        self.assertEqual(expr1, expr2)
        self.assertEqual(hash(expr1), hash(expr2))

    def test_equality_different_type(self):
        expr = create_expression(1)
        self.assertNotEqual(expr, "1")


class TestObserverExpressionSlots(unittest.TestCase):
    """ Check that expressions use __slots__. """

    def test_single_expression(self):
        expr = create_expression(1)
        self.assertFalse(hasattr(expr, "__dict__"))

    def test_series_expression(self):
        expr = create_expression(1).then(create_expression(2))
        self.assertFalse(hasattr(expr, "__dict__"))

    def test_parallel_expression(self):
        expr = create_expression(1) | create_expression(2)
        self.assertFalse(hasattr(expr, "__dict__"))


class TestCompileFromExpr(unittest.TestCase):
    """ Tests for compile_expr. """

    # The complicated pieces are already tested; we just need to double
    # check that "_as_graphs" corresponds to "compile_expr" for a
    # handful of cases.

    def test_compile_expr(self):
        observer1 = 1
        observer2 = 2
        observer3 = 3
        observer4 = 4
        expr1 = create_expression(observer1)
        expr2 = create_expression(observer2)
        expr3 = create_expression(observer3)
        expr4 = create_expression(observer4)

        test_expressions = [
            expr1,
            expr1 | expr2,
            expr1 | expr2 | expr3,
            expr1.then(expr2),
            expr1.then(expr2).then(expr3),
            (expr1.then(expr2)) | (expr3.then(expr4)),
            expr1.list_items(),
            expr1.dict_items(),
            expr1.set_items(),
            expr1.anytrait(notify=False),
            expr1.anytrait(notify=True),
        ]
        for test_expression in test_expressions:
            with self.subTest(expression=test_expression):
                self.assertEqual(
                    expression.compile_expr(test_expression),
                    test_expression._as_graphs(),
                )



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# add_class_trait.py --- Example of mutually-referring classes
#                        using add_class_trait()


# --[Imports]------------------------------------------------------------------
from traits.api import HasTraits, Instance


# --[Code]---------------------------------------------------------------------
# Defining mutually-referring classes using add_class_trait()
class Chicken(HasTraits):
    pass


class Egg(HasTraits):
    created_by = Instance(Chicken)


# Now that 'Egg' is defined, we can add the 'hatched_from' trait to
# solve the mutual-reference problem...

Chicken.add_class_trait("hatched_from", Instance(Egg))



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
This example shows how to set up notifications for changes on a single trait,
for all instances of this class, but only after the object state is set during
instantiation.

The 'post_init' argument suppresses change notifications during instantiation,
but allows subsequent changes to trigger notifications.
"""

from traits.api import HasTraits, Int, observe


class Person(HasTraits):
    age = Int(0)

    @observe("age", post_init=True)
    def notify_age_change(self, event):
        print("age changed from {} to {}".format(event.old, event.new))


person = Person(age=1)  # no output
person.age = 2    # print 'age changed from 1 to 2'



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Tests for Dict items_changed events
"""

import unittest

from traits.api import HasTraits, Dict


class MyClass(HasTraits):
    """ A dummy HasTraits class with a Dict """

    d = Dict({"a": "apple", "b": "banana", "c": "cherry", "d": "durian"})

    def __init__(self, callback):
        "The callback is called with the TraitDictEvent instance"
        self.callback = callback

    def _d_items_changed(self, event):
        if self.callback:
            self.callback(event)


class MyOtherClass(HasTraits):
    """ A dummy HasTraits class with a Dict """

    d = Dict({"a": "apple", "b": "banana", "c": "cherry", "d": "durian"})


class Callback:
    """
    A stateful callback that gets initialized with the values to check for
    """

    def __init__(self, obj, added={}, changed={}, removed={}):
        self.obj = obj
        self.added = added
        self.changed = changed
        self.removed = removed
        self.called = False

    def __call__(self, event):
        self.obj.assertEqual(event.added, self.added)
        self.obj.assertEqual(event.changed, self.changed)
        self.obj.assertEqual(event.removed, self.removed)
        self.called = True


class DictEventTestCase(unittest.TestCase):
    def test_setitem(self):
        # overwriting an existing item
        cb = Callback(self, changed={"c": "cherry"})
        foo = MyClass(cb)
        foo.d["c"] = "coconut"
        self.assertTrue(cb.called)
        # adding a new item
        cb = Callback(self, added={"g": "guava"})
        bar = MyClass(cb)
        bar.d["g"] = "guava"
        self.assertTrue(cb.called)

    def test_delitem(self):
        cb = Callback(self, removed={"b": "banana"})
        foo = MyClass(cb)
        del foo.d["b"]
        self.assertTrue(cb.called)

    def test_clear(self):
        removed = MyClass(None).d.copy()
        cb = Callback(self, removed=removed)
        foo = MyClass(cb)
        foo.d.clear()
        self.assertTrue(cb.called)

    def test_update(self):
        update_dict = {"a": "artichoke", "f": "fig"}
        cb = Callback(self, changed={"a": "apple"}, added={"f": "fig"})
        foo = MyClass(cb)
        foo.d.update(update_dict)
        self.assertTrue(cb.called)

    def test_setdefault(self):
        # Test retrieving an existing value
        cb = Callback(self)
        foo = MyClass(cb)
        self.assertEqual(foo.d.setdefault("a", "dummy"), "apple")
        self.assertFalse(cb.called)

        # Test adding a new value
        cb = Callback(self, added={"f": "fig"})
        bar = MyClass(cb)
        self.assertTrue(bar.d.setdefault("f", "fig") == "fig")
        self.assertTrue(cb.called)

    def test_pop(self):
        # Test popping a non-existent key
        cb = Callback(self)
        foo = MyClass(cb)
        self.assertEqual(foo.d.pop("x", "dummy"), "dummy")
        self.assertFalse(cb.called)

        # Test popping a regular item
        cb = Callback(self, removed={"c": "cherry"})
        bar = MyClass(cb)
        self.assertEqual(bar.d.pop("c"), "cherry")
        self.assertTrue(cb.called)

    def test_popitem(self):
        foo = MyClass(None)
        foo.d.clear()
        foo.d["x"] = "xylophone"
        cb = Callback(self, removed={"x": "xylophone"})
        foo.callback = cb
        self.assertEqual(foo.d.popitem(), ("x", "xylophone"))
        self.assertTrue(cb.called)

    def test_dynamic_listener(self):
        foo = MyOtherClass()
        # Test adding
        func = Callback(self, added={"g": "guava"})
        foo.on_trait_change(func.__call__, "d_items")
        foo.d["g"] = "guava"
        foo.on_trait_change(func.__call__, "d_items", remove=True)
        self.assertTrue(func.called)

        # Test removing
        func2 = Callback(self, removed={"a": "apple"})
        foo.on_trait_change(func2.__call__, "d_items")
        del foo.d["a"]
        foo.on_trait_change(func2.__call__, "d_items", remove=True)
        self.assertTrue(func2.called)

        # Test changing
        func3 = Callback(self, changed={"b": "banana"})
        foo.on_trait_change(func3.__call__, "d_items")
        foo.d["b"] = "broccoli"
        foo.on_trait_change(func3.__call__, "d_items", remove=True)
        self.assertTrue(func3.called)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
This example shows how to set up notifications for a trait that is defined
after the instance is created.

The optional flag suppresses the error that would otherwise occur because the
trait is not found at the time ``observe`` is called. When the observed trait
is defined, the change handler will be aded to the new trait.
"""

from traits.api import HasTraits, Int, observe
from traits.observation.api import trait


class Person(HasTraits):

    @observe(trait("age", optional=True))
    def notify_age_change(self, event):
        print("age changed")


person = Person()
person.add_trait("age", Int())
person.age = 2    # print 'age changed'



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# list_notifier.py -- Example of zero-parameter handlers for an object
#                     containing a list

# --[Imports]------------------------------------------------------------------
from traits.api import HasTraits, List


# --[Code]---------------------------------------------------------------------
class Employee:
    pass


class Department(HasTraits):
    employees = List(Employee)


# --[Example*]-----------------------------------------------------------------
def a_handler():
    print("A handler")


def b_handler():
    print("B handler")


def c_handler():
    print("C handler")


fred = Employee()
mary = Employee()
donna = Employee()

dept = Department(employees=[fred, mary])

# "Old style" name syntax
# a_handler is called only if the list is replaced:
dept.on_trait_change(a_handler, "employees")
# b_handler is called if the membership of the list changes:
dept.on_trait_change(b_handler, "employees_items")

# "New style" name syntax
# c_handler is called if 'employees' or its membership change:
dept.on_trait_change(c_handler, "[employees]")

print("Changing list items")
dept.employees[1] = donna  # Calls B and C
print("Replacing list")
dept.employees = [donna]  # Calls A and C



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# compound.py -- Example of multiple criteria in a trait definition


# --[Imports]------------------------------------------------------------------
from traits.api import HasTraits, Range, Trait, TraitError


# --[Code]---------------------------------------------------------------------
# Shows the definition of a compound trait.


class Die(HasTraits):

    # Define a compound trait definition:
    value = Trait(1, Range(1, 6), "one", "two", "three", "four", "five", "six")


# --[Example*]-----------------------------------------------------------------
# Create a sample Die:
die = Die()

# Try out some sample valid values:
die.value = 3
die.value = "three"
die.value = 5
die.value = "five"

# Now try out some invalid values:
try:
    die.value = 0
except TraitError as excp:
    print(excp)

try:
    die.value = "zero"
except TraitError as excp:
    print(excp)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traits.api import HasTraits, Instance, Str


class Fruit:
    info = Str("good for you")

    def __init__(self, info="good for you", **traits):
        super().__init__(info=info, **traits)


class Orange(Fruit):
    pass


class Pizza:
    pass


def fruit_factory(info, other_stuff):
    return Fruit(info=info)


class TestClass(HasTraits):
    itm = Instance(Fruit)

    itm_args_kw = Instance(Fruit, ('different info',), {'another_trait': 3})
    itm_args = Instance(Fruit, ('different info',))
    itm_kw = Instance(Fruit, {'info': 'different info'})
    itm_factory = Instance(Fruit, fruit_factory)
    itm_factory_args = Instance(Fruit, fruit_factory, ('different info',), )
    itm_factory_args_kw = Instance(
        Fruit,
        fruit_factory,
        ('different info',),
        {'other_stuff': 3},
    )
    itm_factory_kw = Instance(
        Fruit,
        fruit_factory,
        {'other_stuff': 3},
    )


obj = TestClass()
obj.itm = Orange()
obj.itm = Pizza()



# (C) Copyright 2020-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from typing import (
    Any as _Any,
    Optional,
    Callable as _CallableType,
    Dict as _DictType,
)

NoneType: _Any
ConstantTypes: _Any
PythonTypes: _Any
CallableTypes: _Any
TraitTypes: _Any
DefaultValues: _Any


class _InstanceArgs:
    def __init__(self, factory: _Any, args: _Any, kw: _Any) -> None: ...


class Default:
    default_value: _Any = ...

    def __init__(self, func: Optional[_Any] = ..., args: _Any = ...,
                 kw: Optional[_Any] = ...) -> None: ...


def Trait(*value_type: _Any,
          **metadata: _DictType[str, _Any]): ...


def Property(fget: Optional[_CallableType] = ...,
             fset: Optional[_CallableType] = ...,
             fvalidate: Optional[_CallableType] = ...,
             force: bool = ...,
             handler: Optional[_CallableType] = ...,
             trait: Optional[_Any] = ...,
             **metadata: _Any) -> _Any:
    ...


class ForwardProperty:
    metadata: _Any = ...
    validate: _Any = ...
    handler: _Any = ...

    def __init__(self, metadata: _Any,
                 validate: Optional[_Any] = ...,
                 handler: Optional[_Any] = ...) -> None: ...


generic_trait: _Any


def Color(*args: _Any, **metadata: _DictType[str, _Any]): ...


def RGBColor(*args: _Any, **metadata: _DictType[str, _Any]): ...


def Font(*args: _Any, **metadata: _DictType[str, _Any]): ...



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# --(on_trait_change Method Enhancements)--------------------------------------
r"""
on_trait_change Method Enhancements
===================================

In Traits 3.0, the capabilities of the **HasTraits** class's
*on_trait_change* method has been greatly enhanced with the addition of a new
*extended* trait name syntax for specifying the name of the trait the
notification handler applies to.

Previously, the trait name argument to the *on_trait_change* method could only
be one of the following:

Omitted, None, or 'anytrait'
    The notification handler applies to any trait on the object.

name
    The notification handler applies to the trait on the object called *name*.

[name1, ..., namen]
    The notification handler applies to each of the traits on the object with
    the specified names.

In Traits 3.0, all of these forms are still supported, but now the syntax for
specifying *name* has been expanded to allow a much broader set of traits that
are *reachable* from the object the *on_trait_change* method is applied to.

New *name* Parameter Syntax
---------------------------

In Traits 3.0, the *name* parameter, in addition to being omitted, None or
*anytrait*, can now be a single *xname* or a list of *xname* names, where
an *xname* is an extended name of the form::

    xname2['.'xname2]*

An *xname2* is of the form::

   (xname3 | '['xname3[','xname3]*']) ['*']

An *xname3* is of the form::

     xname | ['+'|'-'][name] | name['?' | ('+'|'-')[name]]

A *name* is any valid Python attribute name. The semantic meaning of this
notation is as follows:

[item, item, ..., item]
    A list which matches any of the specified items. Note that at the topmost
    level, the surrounding square brackets are optional.

name?
    If the current object does not have an attribute called *name*, the
    reference can be ignored. If the '?' character is omitted, the current
    object must have a trait called *name*, otherwise an exception will be
    raised.

prefix+
    Matches any trait on the current object whose name begins with *prefix*.

\+metadata_name
    Matches any trait on the current object having *metadata_name* metadata.

\-metadata_name
    Matches any trait on the current object which does not have *metadata_name*
    metadata.

prefix+metadata_name
    Matches any trait on the current object whose name begins with *prefix* and
    which has *metadata_name* metadata.

prefix-metadata_name
    Matches any trait on the current object whose name begins with *prefix* and
    which does not have *metadata_name* metadata.

\+
    Matches all traits on the current object.

pattern*
    Matches object graphs where *pattern* occurs one or more times (useful for
    setting up listeners on recursive data structures like trees or linked
    lists).

Name Syntax Examples
--------------------

Some examples of valid names and their meaning are as follows:

'foo,bar,baz'
    Listen for trait changes to *object.foo*, *object.bar*, and *object.baz*.

['foo','bar','baz']
    Equivalent to 'foo,bar,baz', but may be more useful in cases where the
    individual items are computed.

'foo.bar.baz'
    Listen for trait changes to *object.foo.bar.baz*.

'foo.[bar,baz]'
    Listen for trait changes to *object.foo.bar* and *object.foo.baz*.

'([left,right]).name'
    Listen for trait changes to the *name* trait of each node of a tree having
    *left* and *right* links to other tree nodes, and where *object* is the
    root node of the tree.

'+dirty'
    Listen for trait changes on any trait in the *object* which has the 'dirty'
    metadata set.

'foo.+dirty'
    Listen for trait changes on any trait in *object.foo* which has the 'dirty'
    metadata set.

'foo.[bar,-dirty]'
    Listen for trait changes on *object.foo.bar* or any trait on *object.foo*
    which does not have 'dirty' metadata set.

Additional Semantic Rules
-------------------------

Note that any of the intermediate (i.e., non-final) links in a pattern can be
traits of type **Instance**, **List** or **Dict**. In the case of **List** and
**Dict** traits, the subsequent portion of the pattern is applied to each item
in the list, or value in the dictionary.

For example, if *self.children* is a list, 'children.name' listens for trait
changes to the *name* trait for each item in the *self.children* list.

Also note that items added to or removed from a list or dictionary in the
pattern will cause the *handler* routine to be invoked as well, since this is
treated as an *implied* change to the item's trait being monitored.

Notification Handler Signatures
-------------------------------

The signature of the *handler* supplied also has an effect on how changes to
intermediate traits are processed. The five valid handler signatures are:

1. handler()
2. handler(new)
3. handler(name,new)
4. handler(object,name,new)
5. handler(object,name,old,new)

For signatures 1, 4 and 5, any change to any element of a path being listened
to invokes the handler with information about the particular element that was
modified (e.g., if the item being  monitored is 'foo.bar.baz', a change to
'bar' will call *handler* with the following information:

object
    object.foo
name
    bar
old
    old value for object.foo.bar
new
    new value for object.foo.bar

If one of the intermediate links is a **List** or **Dict**, the call to
*handler* may report an *_items* changed event. If in the previous
example, *bar* is a **List**, and a new item is added to *bar*, then
the information passed to *handler* would be:

object object.foo name bar_items old **Undefined** new **TraitListEvent**
    whose *added* trait contains the new item added to *bar*.

For signatures 2 and 3, the *handler* does not receive enough information to
discern between a change to the final trait being listened to and a change to
an intermediate link. In this case, the event dispatcher will attempt to map a
change to an intermediate link to its effective change on the final trait.
This only works if all of the intermediate links are single values (such as an
**Instance** or **Any** trait) and not **Lists** or **Dicts**. If the modified
intermediate trait or any subsequent intermediate trait preceding the final
trait is a **List** or **Dict**, then a **TraitError** is raised, since the
effective value for the final trait cannot in general be resolved
unambiguously.

Handler signature 1 also has the special characteristic that if a final trait
is a **List** or **Dict**, it will automatically handle *_items* changed
events for the final trait as well. This can be useful in cases where the
*handler* only needs to know that some aspect of the final trait has been
changed. For all other *handler* signatures, you must explicitly specify the
*xxx_items* trait if you want to be notified of changes to any of the items of
the *xxx* trait.

Backward Compatibility
----------------------

The new extended trait name support in Traits 3.0 has one slight semantic
difference with the pre-Traits 3.0 *on_trait_change* method.

Prior to Traits 3.0, it was necessary to make two separate calls to
*on_trait_change* in order to set up listeners on a **List** or **Dict**
trait's value and the contents of its value, as shown in the following
example::

    class Department(HasTraits):

        employees = List(Employee)

        ...

    a_department.on_trait_change(some_listener, 'employees')
    a_department.on_trait_change(some_listener_items, 'employees_items')

In Traits 3.0, this is still the case if the *some_listener* function has one
or more arguments. However, if it has no arguments, the *on_trait_change*
method will automatically call the function either when the trait's value or
its value's contents change. So in Traits 3.0 it is only necessary to write::

    a_department.on_trait_change(some_listener, 'employees')

if the *some_listener* (and *some_listener_items*) function has no arguments.

The net effect of this difference is that code written prior to Traits 3.0
could set up two listeners (e.g. *some_listener* and *some_listener_items*, as
in the example), and then have *both* methods called when the contents of the
trait are modified if the *some_listener* method takes no arguments. Since no
data is passed to the *some_listener* function, there is probably no harm in
doing this, but it does create unnecessary notification handler calls.

As a result, to avoid creating this unwanted overhead in existing code, the
*on_trait_change* method applies pre-Traits 3.0 semantics to all simple names
passed to it (e.g. 'employees'). If you are writing new code and want to
take advantage of the new Traits 3.0 *on_trait_change* semantics for a simple
trait name, you will need to modify the name to use some recognizable aspect of
the new extended trait name syntax.

For example, either of the following lines would cause *new style* semantics to
be applied to the *employees* trait::

    a_department.on_trait_change(some_listener, ' employees')

    a_department.on_trait_change(some_listener, '[employees]')

A Complete Example
------------------

Refer to the code tabs of this lesson for a complete example using
*on_trait_change* with an extended trait name. In particular, check near the
bottom of the **Example** tab for the code that sets up an extended trait
change notification handler using *on_trait_change*.
"""

from traits.api import HasTraits, Int, List, Str


# --[Employee Class]-----------------------------------------------------------
class Employee(HasTraits):

    # The name of the employee:
    name = Str

    # The number of sick days they have taken this year:
    sick_days = Int


# --[Department Class]---------------------------------------------------------
class Department(HasTraits):

    # The name of the department:
    name = Str

    # The employees in the department:
    employees = List(Employee)


# --[Corporation Class]--------------------------------------------------------
class Corporation(HasTraits):

    # The name of the corporation:
    name = Str

    # The departments within the corporation:
    departments = List(Department)


# --[Example*]-----------------------------------------------------------------
# Create some sample employees:
millie = Employee(name="Millie", sick_days=2)
ralph = Employee(name="Ralph", sick_days=3)
tom = Employee(name="Tom", sick_days=1)
slick = Employee(name="Slick", sick_days=16)
marcelle = Employee(name="Marcelle", sick_days=7)
reggie = Employee(name="Reggie", sick_days=11)
dave = Employee(name="Dave", sick_days=0)
bob = Employee(name="Bob", sick_days=1)
alphonse = Employee(name="Alphonse", sick_days=5)

# Create some sample departments:
accounting = Department(name="accounting", employees=[millie, ralph, tom])

sales = Department(name="Sales", employees=[slick, marcelle, reggie])

development = Department(name="Development", employees=[dave, bob, alphonse])

# Create a sample corporation:
acme = Corporation(
    name="Acme, Inc.", departments=[accounting, sales, development]
)


# Define a corporate 'whistle blower' function:
def sick_again(object, name, old, new):
    print(
        "%s just took sick day number %d for this year!" % (object.name, new)
    )


# Set up the function as a listener:
acme.on_trait_change(sick_again, "departments.employees.sick_days")

# Now let's try it out:
slick.sick_days += 1
reggie.sick_days += 1



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traits.api import Float, HasTraits, Interface, provides, Str, Tuple


class IHasName(Interface):
    name = Str()


@provides(IHasName)
class NamedColor(HasTraits):
    name = Str()

    rgb = Tuple(Float, Float, Float)


named_color = NamedColor(name="green", rgb=(0.0, 1.0, 0.0))



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

#  fixme:
#  - Get custom tree view images.
#  - Write a program to create a directory structure from a lesson plan file.

""" Script to run the tutorial.
"""


import os
import sys

from traitsui.extras.demo import demo

# Correct program usage information:
usage = """
Correct usage is: tutor.py [root_dir]
where:
    root_dir = Path to root of the tutorial tree

If omitted, 'root_dir' defaults to the current directory."""


def main(root_dir):
    # Create a tutor and display the tutorial:
    path, name = os.path.splitext(root_dir)
    demo(dir_name=root_dir, title='Traits Demos')


if __name__ == "__main__":

    # Validate the command line arguments:
    if len(sys.argv) > 2:
        print(usage)
        sys.exit(1)

    # Determine the root path to use for the tutorial files:
    if len(sys.argv) == 2:
        root_dir = sys.argv[1]
    else:
        root_dir = os.getcwd()

    main(root_dir)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# deferring_notification.py -- Example of notification with deferring

# --[Imports]------------------------------------------------------------------
from traits.api import HasTraits, Instance, PrototypedFrom, Str


# --[Code]---------------------------------------------------------------------
class Parent(HasTraits):

    first_name = Str
    last_name = Str

    def _last_name_changed(self, new):
        print("Parent's last name changed to %s." % new)


class Child(HasTraits):

    father = Instance(Parent)
    first_name = Str
    last_name = PrototypedFrom("father")

    def _last_name_changed(self, new):
        print("Child's last name changed to %s." % new)


# --[Example*]-----------------------------------------------------------------
dad = Parent(first_name="William", last_name="Chase")
# Output: Parent's last name changed to Chase.

son = Child(first_name="John", father=dad)
# Output: Child's last name changed to Chase.

# Change Parent's last_name
dad.last_name = "Jones"
# Output: Parent's last name changed to Jones.
#         Child's last name changed to Jones.

# Override Child's last_name
son.last_name = "Thomas"
# Output Child's last name changed to Thomas.

# Change Parent's last_name; Child's is not affected.
dad.last_name = "Riley"
# Output: Parent's last name changed to Riley.

# Reset Child's last_name
del son.last_name
# Output: Child's last name changed to Riley.

# Change to Parent now affects Child.
dad.last_name = "Simmons"
# Output: Parent's last name changed to Simmons.
#         Child's last name changed to Simmons.



# (C) Copyright 2020-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import abc
from .adaptation.adaptation_error import AdaptationError as AdaptationError
from .constants import DefaultValue as DefaultValue, TraitKind as TraitKind
from .ctrait import CTrait as CTrait, __newobj__ as __newobj__
from .ctraits import CHasTraits as CHasTraits
from .trait_base import SequenceTypes as SequenceTypes, TraitsCache as TraitsCache, Undefined as Undefined, is_none as is_none, not_event as not_event, not_false as not_false
from .trait_converters import check_trait as check_trait, mapped_trait_for as mapped_trait_for, trait_for as trait_for
from .trait_errors import TraitError as TraitError
from .trait_notifiers import ExtendedTraitChangeNotifyWrapper as ExtendedTraitChangeNotifyWrapper, FastUITraitChangeNotifyWrapper as FastUITraitChangeNotifyWrapper, NewTraitChangeNotifyWrapper as NewTraitChangeNotifyWrapper, StaticAnytraitChangeNotifyWrapper as StaticAnytraitChangeNotifyWrapper, StaticTraitChangeNotifyWrapper as StaticTraitChangeNotifyWrapper, TraitChangeNotifyWrapper as TraitChangeNotifyWrapper
from .trait_types import Any as Any, Bool as Bool, Disallow as Disallow, Event as Event, Python as Python
from .traits import ForwardProperty as ForwardProperty, Property as Property, Trait as Trait, generic_trait as generic_trait
from .util.deprecated import deprecated as deprecated
from typing import Optional, Any as _Any

CHECK_INTERFACES: int

class AbstractViewElement(abc.ABC): ...

WrapperTypes: _Any
BaseTraits: str
ClassTraits: str
PrefixTraits: str
ListenerTraits: str
ViewTraits: str
InstanceTraits: str
DefaultTraitsView: str
CantHaveDefaultValue: _Any
DeferredCopy: _Any
extended_trait_pat: _Any
any_trait: _Any

def is_unbound_method_type(method: _Any): ...
def get_delegate_pattern(name: _Any, trait: _Any): ...

class _SimpleTest:
    value: _Any = ...
    def __init__(self, value: _Any) -> None: ...
    def __call__(self, test: _Any): ...

class MetaHasTraits(type):
    def __new__(cls, class_name: _Any, bases: _Any, class_dict: _Any): ...
    @classmethod
    def add_listener(cls, listener: _Any, class_name: str = ...) -> None: ...
    @classmethod
    def remove_listener(cls, listener: _Any, class_name: str = ...) -> None: ...

def update_traits_class_dict(class_name: _Any, bases: _Any, class_dict: _Any): ...
def migrate_property(name: _Any, property: _Any, property_info: _Any, class_dict: _Any): ...
def observe(expression: _Any, post_init: bool = ..., dispatch: str = ...): ...
def on_trait_change(name: _Any, post_init: bool = ..., dispatch: str = ...): ...
def cached_property(function: _Any): ...
def property_depends_on(dependency: _Any, settable: bool = ..., flushable: bool = ...): ...
def weak_arg(arg: _Any): ...

class HasTraits(CHasTraits, metaclass=MetaHasTraits):
    _traits_cache__: _Any = ...
    wrappers: _Any = ...
    trait_added: _Any = ...
    trait_modified: _Any = ...
    @classmethod
    def add_class_trait(cls, name: _Any, *trait: _Any) -> None: ...
    @classmethod
    def set_trait_dispatch_handler(cls, name: _Any, klass: _Any, override: bool = ...) -> None: ...
    @classmethod
    def trait_subclasses(cls, all: bool = ...): ...
    def has_traits_interface(self, *interfaces: _Any): ...
    def __reduce_ex__(self, protocol: _Any): ...
    def trait_get(self, *names: _Any, **metadata: _Any): ...
    def get(self, *names: _Any, **metadata: _Any): ...
    def trait_set(self, trait_change_notify: bool = ..., **traits: _Any): ...
    def set(self, trait_change_notify: bool = ..., **traits: _Any): ...
    def trait_setq(self, **traits: _Any): ...
    def reset_traits(self, traits: Optional[_Any] = ..., **metadata: _Any): ...
    def copyable_trait_names(self, **metadata: _Any): ...
    def all_trait_names(self): ...
    def __dir__(self): ...
    def copy_traits(self, other: _Any, traits: Optional[_Any] = ..., memo: Optional[_Any] = ..., copy: Optional[_Any] = ..., **metadata: _Any): ...
    def clone_traits(self, traits: Optional[_Any] = ..., memo: Optional[_Any] = ..., copy: Optional[_Any] = ..., **metadata: _Any): ...
    def __deepcopy__(self, memo: _Any): ...
    def edit_traits(self, view: Optional[_Any] = ..., parent: Optional[_Any] = ..., kind: Optional[_Any] = ..., context: Optional[_Any] = ..., handler: Optional[_Any] = ..., id: str = ..., scrollable: Optional[_Any] = ..., **args: _Any): ...
    def trait_context(self): ...
    def trait_view(self, name: Optional[_Any] = ..., view_element: Optional[_Any] = ...): ...
    @classmethod
    def class_trait_view(cls, name: Optional[_Any] = ..., view_element: Optional[_Any] = ...): ...
    def default_traits_view(self): ...
    @classmethod
    def class_default_traits_view(cls): ...
    def trait_views(self, klass: Optional[_Any] = ...): ...
    def trait_view_elements(self): ...
    @classmethod
    def class_trait_view_elements(cls): ...
    def configure_traits(self, filename: Optional[_Any] = ..., view: Optional[_Any] = ..., kind: Optional[str] = ..., edit: bool = ..., context: Optional[_Any] = ..., handler: Optional[_Any] = ..., id: str = ..., scrollable: Optional[_Any] = ..., **args: _Any): ...
    def editable_traits(self): ...
    @classmethod
    def class_editable_traits(cls): ...
    def visible_traits(self): ...
    @classmethod
    def class_visible_traits(cls): ...
    def print_traits(self, show_help: bool = ..., **metadata: _Any) -> None: ...
    def observe(self, handler: _Any, expression: _Any, remove: bool = ..., dispatch: str = ...) -> None: ...
    def on_trait_change(self, handler: _Any, name: Optional[_Any] = ..., remove: bool = ..., dispatch: str = ..., priority: bool = ..., deferred: bool = ..., target: Optional[_Any] = ...) -> None: ...
    on_trait_event: _Any = ...
    def sync_trait(self, trait_name: _Any, object: _Any, alias: Optional[_Any] = ..., mutual: bool = ..., remove: bool = ...): ...
    def add_trait(self, name: _Any, *trait: _Any) -> None: ...
    def remove_trait(self, name: _Any): ...
    def trait(self, name: str, force: bool = ..., copy: bool = ...): ...
    def base_trait(self, name: _Any): ...
    def validate_trait(self, name: _Any, value: _Any): ...
    def traits(self, **metadata: _Any): ...
    @classmethod
    def class_traits(cls, **metadata: _Any): ...
    def trait_names(self, **metadata: _Any): ...
    @classmethod
    def class_trait_names(cls, **metadata: _Any): ...
    def __prefix_trait__(self, name: _Any, is_set: _Any): ...
    def add_trait_listener(self, object: _Any, prefix: str = ...) -> None: ...
    def remove_trait_listener(self, object: _Any, prefix: str = ...) -> None: ...

class HasStrictTraits(HasTraits): ...

class HasRequiredTraits(HasStrictTraits):
    def __init__(self, **traits: _Any) -> None: ...

class HasPrivateTraits(HasTraits):
    __: _Any = ...

class ABCMetaHasTraits(abc.ABCMeta, MetaHasTraits): ...
class ABCHasTraits(HasTraits, metaclass=ABCMetaHasTraits): ...
class ABCHasStrictTraits(ABCHasTraits): ...

class SingletonHasTraits(HasTraits):
    def __new__(cls, *args: _Any, **traits: _Any): ...

class SingletonHasStrictTraits(HasStrictTraits):
    def __new__(cls, *args: _Any, **traits: _Any): ...

class SingletonHasPrivateTraits(HasPrivateTraits):
    def __new__(cls, *args: _Any, **traits: _Any): ...

class Vetoable(HasStrictTraits):
    veto: _Any = ...

VetoableEvent: _Any

class MetaInterface(ABCMetaHasTraits):
    def __call__(self, adaptee: _Any, default: _Any = ...): ...

class Interface(HasTraits, metaclass=MetaInterface): ...

def provides(*protocols: _Any): ...
def isinterface(klass: _Any): ...

class ISerializable(Interface): ...



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# --[Imports]------------------------------------------------------------------
from traits.api import HasTraits, Instance


# --[Code]---------------------------------------------------------------------
# Shows the incorrect way of defining mutually-referring classes.
try:

    class Chicken(HasTraits):

        # Won't work: 'Egg' not defined yet:
        hatched_from = Instance(Egg)

    class Egg(HasTraits):

        # If we move this class to the top, then this line won't work, because
        # 'Chicken' won't be defined yet:
        created_by = Instance(Chicken)


except NameError as excp:
    print(excp)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from functools import lru_cache

from traits.observation import _generated_parser
import traits.observation.expression as expression_module


_LARK_PARSER = _generated_parser.Lark_StandAlone()


def _handle_series(trees, notify):
    """ Handle an expression of the form "a.b" or "a:b".

    Parameters
    ----------
    trees : list of lark.tree.Tree
        The children tree for the "series" rule. It should always
        contain exactly three items.
    notify : bool
        True if the final target should notify, else False.

    Returns
    -------
    expression : ObserverExpression
    """
    left, connector, right = trees
    notify_left = connector.data == "notify"
    return _handle_tree(left, notify_left).then(_handle_tree(right, notify))


def _handle_parallel(trees, notify):
    """ Handle an expression of the form "a, b".

    Parameters
    ----------
    trees : list of lark.tree.Tree
        The children tree for the "parallel" rule. It should always
        contain exactly two items.
    notify : bool
        True if the final target should notify, else False.

    Returns
    -------
    expression : ObserverExpression
    """
    left, right = trees
    return _handle_tree(left, notify) | _handle_tree(right, notify)


def _handle_trait(trees, notify):
    """ Handle an element for a named trait.

    Parameters
    ----------
    trees : list of lark.tree.Tree
        The children tree for the "trait" rule.
        It contains only one item.
    notify : bool
        True if the final target should notify, else False.

    Returns
    -------
    expression : ObserverExpression
    """
    token, = trees
    name = token.value
    return expression_module.trait(name, notify=notify)


def _handle_anytrait(trees, notify):
    """ Handle an anytrait element.

    Parameters
    ----------
    trees : list of lark.tree.Tree
        The children tree for the "trait" rule. This should be empty.
    notify : bool
        True if the final target should notify, else False.

    Returns
    -------
    expression : ObserverExpression
    """
    return expression_module.anytrait(notify=notify)


def _handle_metadata(trees, notify):
    """ Handle an element for filtering existing metadata.

    Parameters
    ----------
    trees : list of lark.tree.Tree
        The children tree for the "metadata" rule.
        It contains only one item.
    notify : bool
        True if the final target should notify, else False.

    Returns
    -------
    expression : ObserverExpression
    """
    token, = trees
    metadata_name = token.value
    return expression_module.metadata(metadata_name, notify=notify)


def _handle_items(trees, notify):
    """ Handle keyword "items".

    Parameters
    ----------
    trees : list of lark.tree.Tree
        The children tree for the "items" rule.
        It should be empty.
    notify : bool
        True if the final target should notify, else False.

    Returns
    -------
    expression : ObserverExpression
    """
    if trees:
        # Nothing should be wrapped in items
        raise ValueError("Unexpected tree: {!r}".format(trees))

    return (
        expression_module.trait("items", notify=notify, optional=True)
        | expression_module.dict_items(notify=notify, optional=True)
        | expression_module.list_items(notify=notify, optional=True)
        | expression_module.set_items(notify=notify, optional=True)
    )


def _handle_tree(tree, notify):
    """ Handle a tree using the specified rule.

    Parameters
    ----------
    tree : lark.tree.Tree
        Tree to be converted to an ObserverExpression.
    notify : bool
        True if the final target should notify, else False.

    Returns
    -------
    expression: ObserverExpression
    """
    # All handlers must be callable
    # with the signature (list of Tree, notify)
    handlers = {
        "series": _handle_series,
        "series_terminal": _handle_series,
        "parallel": _handle_parallel,
        "parallel_terminal": _handle_parallel,
        "trait": _handle_trait,
        "metadata": _handle_metadata,
        "items": _handle_items,
        "anytrait": _handle_anytrait,
    }
    return handlers[tree.data](tree.children, notify)


@lru_cache(maxsize=expression_module._OBSERVER_EXPRESSION_CACHE_MAXSIZE)
def parse(text):
    """ Top-level function for parsing user's text to an ObserverExpression.

    Parameters
    ----------
    text : str
        Text to be parsed.

    Returns
    -------
    expression : ObserverExpression

    Raises
    ------
    ValueError
        If the text is not a valid observe expression.
    """
    try:
        tree = _LARK_PARSER.parse(text)
    except _generated_parser.LarkError as parser_exception:
        raise ValueError(f"Invalid expression: {text!r}") from parser_exception

    return _handle_tree(tree, notify=True)


@lru_cache(maxsize=expression_module._OBSERVER_EXPRESSION_CACHE_MAXSIZE)
def compile_str(text):
    """ Compile a mini-language string to a list of ObserverGraph objects.

    Parameters
    ----------
    text : str
        Text to be parsed.

    Returns
    -------
    list of ObserverGraph
    """
    return expression_module.compile_expr(parse(text))



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.api import HasTraits, TraitError
from traits.trait_types import _NoneTrait


class A(HasTraits):
    none_atr = _NoneTrait(default_value=None)


class TestCaseNoneTrait(unittest.TestCase):
    def test_none(self):
        obj = A()
        self.assertIsNone(obj.none_atr)

    def test_assign_non_none(self):
        with self.assertRaises(TraitError):
            A(none_atr=5)

    def test_default_value_not_none(self):
        with self.assertRaises(ValueError):
            class TestClass(HasTraits):
                none_trait = _NoneTrait(default_value=[])



# (C) Copyright 2020-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


from pathlib import Path
from unittest import TestCase

from traits.testing.optional_dependencies import (
    pkg_resources,
    requires_mypy,
    requires_numpy_typing,
    requires_pkg_resources,
)
from traits_stubs_tests.util import MypyAssertions


@requires_pkg_resources
@requires_mypy
class TestAnnotations(TestCase, MypyAssertions):
    def test_all(self, filename_suffix=""):
        """ Run mypy for all files contained in traits_stubs_tests/examples
        directory.

        Lines with expected errors are marked inside these files.
        Any mismatch will raise an assertion error.

        Parameters
        ----------
        filename_suffix: str
            Optional filename suffix filter.
        """
        examples_dir = Path(pkg_resources.resource_filename(
            'traits_stubs_tests', 'examples'))

        for file_path in examples_dir.glob("*{}.py".format(filename_suffix)):
            with self.subTest(file_path=file_path):
                self.assertRaisesMypyError(file_path)

    @requires_numpy_typing
    def test_numpy_examples(self):
        """ Run mypy for files contained in traits_stubs_tests/numpy_examples
        directory.

        Lines with expected errors are marked inside these files.
        Any mismatch will raise an assertion error.

        Parameters
        ----------
        filename_suffix: str
            Optional filename suffix filter.
        """
        examples_dir = Path(pkg_resources.resource_filename(
            'traits_stubs_tests', 'numpy_examples'))

        for file_path in examples_dir.glob("*.py"):
            with self.subTest(file_path=file_path):
                self.assertRaisesMypyError(file_path)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.util.weakiddict import WeakIDDict, WeakIDKeyDict


class AllTheSame(object):
    def __hash__(self):
        return 42

    def __eq__(self, other):
        return isinstance(other, type(self))

    def __ne__(self, other):
        return not self.__eq__(other)


class WeakreffableInt(object):
    def __init__(self, value):
        self.value = value

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        if isinstance(other, int):
            return self.value == other
        else:
            return self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)


class TestWeakIDDict(unittest.TestCase):

    def test_weak_keys(self):
        wd = WeakIDKeyDict()

        keep = []
        dont_keep = []
        values = list(range(10))
        for n, i in enumerate(values, 1):
            key = AllTheSame()
            if not (i % 2):
                keep.append(key)
            else:
                dont_keep.append(key)
            wd[key] = i
            del key
            # No keys or values have been deleted, yet.
            self.assertEqual(len(wd), n)

        # Delete half of the keys.
        self.assertEqual(len(wd), 10)
        del dont_keep
        self.assertEqual(len(wd), 5)

        # Check the remaining values.
        self.assertCountEqual(list(wd.values()), list(range(0, 10, 2)))
        self.assertEqual([wd[k] for k in keep], list(range(0, 10, 2)))

        # Check the remaining keys.
        self.assertCountEqual([id(k) for k in wd.keys()], [id(k) for k in wd])
        self.assertCountEqual(
            [id(k) for k in wd.keys()], [id(k) for k in keep]
        )

    def test_weak_keys_values(self):
        wd = WeakIDDict()

        keep = []
        dont_keep = []
        values = list(map(WeakreffableInt, range(10)))
        for n, i in enumerate(values, 1):
            key = AllTheSame()
            if not (i.value % 2):
                keep.append(key)
            else:
                dont_keep.append(key)
            wd[key] = i
            del key
            # No keys or values have been deleted, yet.
            self.assertEqual(len(wd), n)

        # Delete half of the keys.
        self.assertEqual(len(wd), 10)
        del dont_keep
        self.assertEqual(len(wd), 5)

        # Check the remaining values.
        self.assertCountEqual(
            list(wd.values()), list(map(WeakreffableInt, [0, 2, 4, 6, 8]))
        )
        self.assertEqual(
            [wd[k] for k in keep], list(map(WeakreffableInt, [0, 2, 4, 6, 8]))
        )

        # Check the remaining keys.
        self.assertCountEqual([id(k) for k in wd.keys()], [id(k) for k in wd])
        self.assertCountEqual(
            [id(k) for k in wd.keys()], [id(k) for k in keep]
        )

        # Delete the weak values progressively and ensure that the
        # corresponding entries disappear.
        del values[0:2]
        self.assertEqual(len(wd), 4)
        del values[0:2]
        self.assertEqual(len(wd), 3)
        del values[0:2]
        self.assertEqual(len(wd), 2)
        del values[0:2]
        self.assertEqual(len(wd), 1)
        del values[0:2]
        self.assertEqual(len(wd), 0)

    def test_weak_id_dict_str_representation(self):
        """ test string representation of the WeakIDDict class. """
        weak_id_dict = WeakIDDict()
        desired_repr = "<WeakIDDict at 0x{0:x}>".format(id(weak_id_dict))
        self.assertEqual(desired_repr, str(weak_id_dict))
        self.assertEqual(desired_repr, repr(weak_id_dict))

    def test_weak_id_key_dict_str_representation(self):
        """ test string representation of the WeakIDKeyDict class. """
        weak_id_key_dict = WeakIDKeyDict()
        desired_repr = f"<WeakIDKeyDict at 0x{id(weak_id_key_dict):x}>"
        self.assertEqual(desired_repr, str(weak_id_key_dict))
        self.assertEqual(desired_repr, repr(weak_id_key_dict))



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# static_notification.py --- Example of static attribute notification

# --[Imports]------------------------------------------------------------------
from traits.api import HasTraits, Float


# --[Code]---------------------------------------------------------------------
class Person(HasTraits):
    weight_kg = Float(0.0)
    height_m = Float(1.0)
    bmi = Float(0.0)

    def _weight_kg_changed(self, old, new):
        print("weight_kg changed from %s to %s " % (old, new))
        if self.height_m != 0.0:
            self.bmi = self.weight_kg / (self.height_m ** 2)

    def _anytrait_changed(self, name, old, new):
        print("The %s trait changed from %s to %s " % (name, old, new))


# --[Example*]-----------------------------------------------------------------
bob = Person()
bob.height_m = 1.75
# Output: The height_m trait changed from 1.0 to 1.75
bob.weight_kg = 100.0
# Output:
# The weight_kg trait changed from 0.0 to 100.0
# weight_kg changed from 0.0 to 100.0
# The bmi trait changed from 0.0 to 32.6530612245



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# --(Delegation Fixes and Improvements)----------------------------------------
"""
Delegation Fixes and Improvements
=================================

In previous versions of Traits there were a number of problems (i.e. bugs) in
the delegation support that made delegation virtually unusable in some
situations.

As a result, one of the primary goals of Traits 3.0 was to fix all known
problems with delegation, thus allowing it to reclaim its role as one of the
five keys pillars of the Traits package (those pillars being *initialization*,
*validation*, *notification*, *delegation* and *visualization*).

We are happy to report that not only have all known bugs been fixed, but a
previously unsupported, though often requested, feature has been added as well.

Delegation Event Notification
-----------------------------

Previously, many Traits users implicitly assumed that changes made to a
*delegatee* trait would generate a change notification on any *delegater* trait
(no matter how many such traits there might be). Unfortunately, this was not
the case.

However, starting with Traits 3.0, this feature has now been implemented. An
example of what this means is shown below::

    class Parent(HasTraits):

        first_name = Str
        last_name  = Str

    class Child(HasTraits):

        mother = Instance(Parent)
        father = Instance(Parent)

        first_name = Str
        last_name  = Delegate('father')

In this example, we've created two classes, **Parent** and **Child**, and the
value of the **Child** class's *last_name* trait delegates its value to its
*father* object's *last_name* trait.

Next, we'll set up a simple set of test objects::

    mom = Parent(first_name='Julia', last_name='Wilson')
    dad = Parent(first_name='William', last_name='Chase')
    son = Child(mother=mom, father=dad, first_name='John')

Finally, we'll set up a notification handler on the *son* object's *last_name*
trait and then try out various combinations of setting both the *father* and
*son* object's *last_name* trait to see in which cases the notification handler
is called::

    def name_changed(name):
        print('Your last name has been changed to %s.' % name)

    # Set up a change notification handler on the son's last name:
    son.on_trait_change(name_changed, 'last_name')

    # This should cause the son's last name to change as well:
    print("Changing dad's last name to Jones.")
    dad.last_name = 'Jones'

    # This change overrides the father's last name for the son:
    print("Changing son's last name to Thomas.")
    son.last_name = 'Thomas'

    # This should no longer have any effect on the son's last name:
    print("Changing dad's last name to Riley.")
    dad.last_name = 'Riley'

    # Son decides to revert his name back to his father's name:
    print("Reverting son's last name.")
    del son.last_name

    # Now changing the father's name should affect the son again:
    print("Changing dad's last name to Simmons.")
    dad.last_name = 'Simmons'

For the actual results of running this code, refer to the **Output** tab. Note
that for each case in which an explicit or implicit change is made to the *son*
object's *last_name* trait, a corresponding call is made to the change
notification handler.

"""

# FIXME - this need to be redone without traitsui

from traits.api import Delegate, HasTraits, Instance, Str


# --[Parent Class]-------------------------------------------------------------
class Parent(HasTraits):

    first_name = Str
    last_name = Str


# --[Child Class]--------------------------------------------------------------
class Child(HasTraits):

    mother = Instance(Parent)
    father = Instance(Parent)

    first_name = Str
    last_name = Delegate("father")


# --[Example*]-----------------------------------------------------------------

mom = Parent(first_name="Julia", last_name="Wilson")
dad = Parent(first_name="William", last_name="Chase")
son = Child(mother=mom, father=dad, first_name="John")


def name_changed(name):
    print("Your last name has been changed to %s." % name)


# Set up a change notification handler on the son's last name:
son.on_trait_change(name_changed, "last_name")

# This should cause the son's last name to change as well:
print("Changing dad's last name to Jones.")
dad.last_name = "Jones"

# This change overrides the father's last name for the son:
print("Changing son's last name to Thomas.")
son.last_name = "Thomas"

# This should no longer have any effect on the son's last name:
print("Changing dad's last name to Riley.")
dad.last_name = "Riley"

# Son decides to revert his name back to his father's name:
print("Reverting son's last name.")
del son.last_name

# Now changing the father's name should affect the son again:
print("Changing dad's last name to Simmons.")
dad.last_name = "Simmons"



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Module for HasTraits and CTrait specific utility functions for observers.
"""

from traits.constants import ComparisonMode, TraitKind
from traits.ctraits import CHasTraits
from traits.observation._observe import add_or_remove_notifiers
from traits.observation.exceptions import NotifierNotFound
from traits.trait_base import Undefined, Uninitialized


#: List of values not to be passed onto the next observers because they are
#: filled values for uninitialized/undefined traits, and no observables
#: can be derived from them.
UNOBSERVABLE_VALUES = [
    Undefined,
    Uninitialized,
    None,
]


def object_has_named_trait(object, name):
    """ Return true if a trait with the given name is defined on the object.

    Parameters
    ----------
    object : any
        Any object
    name : str
        Trait name to look for.

    Returns
    -------
    boolean
    """
    return (
        isinstance(object, CHasTraits) and object._trait(name, 0) is not None
    )


def iter_objects(object, name):
    """ Yield the object referenced by the named trait, unless the value is not
    defined or is one of the few unobservable filled values.

    This function should not introduce side-effect (e.g. evaluating default)
    on the given object.

    Parameters
    ----------
    object : HasTraits
        Object from which the trait value will be obtained.
    name : str
        Name of the trait.

    Yields
    ------
    value : any
        The value of the trait, if it is defined and is not one of those
        skipped values.
    """
    value = object.__dict__.get(name, Undefined)
    if all(value is not skipped for skipped in UNOBSERVABLE_VALUES):
        yield value


def observer_change_handler(event, graph, handler, target, dispatcher):
    """ Maintain downstream notifiers when the trait changes.

    Parameters
    ----------
    event : TraitChangeEvent
        The event that triggers this function.
    graph : ObserverGraph
        Description for the *downstream* observers, i.e. excluding the observer
        that contributed this maintainer function.
    handler : callable
        User handler.
    target : object
        Object seen by the user as the owner of the observer.
    dispatcher : callable
        Callable for dispatching the handler.
    """
    if all(event.old is not skipped for skipped in UNOBSERVABLE_VALUES):
        try:
            add_or_remove_notifiers(
                object=event.old,
                graph=graph,
                handler=handler,
                target=target,
                dispatcher=dispatcher,
                remove=True,
            )
        except NotifierNotFound:
            # The old value could be a default value that does not have any
            # notifier.
            pass

    if all(event.new is not skipped for skipped in UNOBSERVABLE_VALUES):
        add_or_remove_notifiers(
            object=event.new,
            graph=graph,
            handler=handler,
            target=target,
            dispatcher=dispatcher,
            remove=False,
        )


def ctrait_prevent_event(event):
    """ Return true if the CTrait change event should be skipped.

    Parameters
    ----------
    event : TraitChangeEvent

    Returns
    -------
    skipped : bool
        Whether the event should be skipped
    """
    if event.old is Uninitialized:
        return True

    ctrait = event.object.trait(event.name)
    if (ctrait.type == TraitKind.trait.name
            and ctrait.comparison_mode == ComparisonMode.equality):
        try:
            return bool(event.old == event.new)
        except Exception:
            # Maybe do something else about the exception
            # enthought/traits#1230
            pass
    return False



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.api import (
    HasTraits,
    Int,
    List,
    Str,
    TraitError,
)


class Foo(HasTraits):
    num = Int
    bar = Str


class Bar(HasTraits):
    # Default is visible.
    PubT1 = Str
    # Change to invisible.
    PubT2 = Str(visible=False)
    # New behaviour makes private traits invisible.
    PrivT1 = Str(private=True)
    # Force visibility of a private trait.
    PrivT2 = Str(private=True, visible=True)


class FooBar(HasTraits):
    num = Int
    baz = "non-trait class attribute"


class GetTraitTestCase(unittest.TestCase):
    def test_trait_set_bad(self):
        b = Foo(num=23)
        # This should fail before and after #234.
        with self.assertRaises(TraitError):
            b.num = "first"
        self.assertEqual(b.num, 23)

    def test_trait_set_replaced(self):
        b = Foo()
        # Overriding the trait with a new type should work.
        b.add_trait("num", Str())
        b.num = "first"
        self.assertEqual(b.num, "first")

    def test_trait_set_replaced_and_check(self):
        b = Foo()
        b.add_trait("num", Str())
        b.num = "first"
        self.assertEqual(b.num, "first")

        # Check that the "traits" call picks up the new instance trait. (See
        # #234.)
        self.assertEqual(b.trait("num"), b.traits()["num"])

    def test_trait_names_returned_by_visible_traits(self):
        b = Bar()
        self.assertEqual(
            sorted(b.visible_traits()), sorted(["PubT1", "PrivT2"])
        )

    def test_dir(self):
        b = FooBar()
        names = dir(b)
        self.assertIn("baz", names)
        self.assertIn("num", names)
        self.assertIn("edit_traits", names)

        # Issue 925: _notifiers not shown in dir()
        self.assertIn("_notifiers", names)

        # Ensure no duplicates
        self.assertEqual(len(set(names)), len(names))

    def test_trait_name_with_list_items(self):
        # Dynamically added a List trait causes an "_items" to be added, but
        # that should not be reported by HasTraits.traits

        class Base(HasTraits):
            pass

        a = Base()
        # This adds an instance trait for "pins_items" as well.
        a.add_trait("pins", List())

        self.assertIn("pins", a.traits())
        self.assertNotIn("pins_items", a.traits())

    def test_trait_name_with_items(self):
        # Dynamically added traits with name "*_items" is also skipped.

        class Base(HasTraits):
            pass

        a = Base()
        a.add_trait("good_items", Str())

        # If enthought/traits#1329 is fixed, then the following assertion
        # should fail.
        self.assertNotIn("good_items", a.traits())



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Tests for the Date trait type.
"""

import datetime
import unittest

from traits.testing.optional_dependencies import requires_traitsui, traitsui

from traits.api import Date, HasStrictTraits, TraitError


#: Unix epoch date.
UNIX_EPOCH = datetime.date(1970, 1, 1)

#: Windows NT epoch
NT_EPOCH = datetime.date(1600, 1, 1)


class HasDateTraits(HasStrictTraits):
    #: Simple case - no default, no parameters, no metadata
    simple_date = Date()

    #: Date with default
    epoch = Date(UNIX_EPOCH)

    #: Date with default provided via keyword.
    alternative_epoch = Date(default_value=NT_EPOCH)

    #: Datetime instances prohibited
    datetime_prohibited = Date(allow_datetime=False)

    #: Datetime instances allowed
    datetime_allowed = Date(allow_datetime=True)

    #: None prohibited
    none_prohibited = Date(allow_none=False)

    #: None allowed
    none_allowed = Date(allow_none=True)

    #: Strictly a non-None non-datetime date
    strict = Date(allow_datetime=False, allow_none=False)


class TestDate(unittest.TestCase):
    def test_default(self):
        obj = HasDateTraits()
        self.assertEqual(obj.simple_date, None)
        self.assertEqual(obj.epoch, UNIX_EPOCH)
        self.assertEqual(obj.alternative_epoch, NT_EPOCH)

    def test_assign_date(self):
        test_date = datetime.date(1975, 2, 13)
        obj = HasDateTraits()
        obj.simple_date = test_date
        self.assertEqual(obj.simple_date, test_date)

    def test_assign_non_date(self):
        obj = HasDateTraits()
        with self.assertRaises(TraitError) as exception_context:
            obj.simple_date = "1975-2-13"
        message = str(exception_context.exception)
        self.assertIn("must be a date or None, but", message)

    def test_assign_none_with_allow_none_not_given(self):
        obj = HasDateTraits(simple_date=UNIX_EPOCH)
        self.assertIsNotNone(obj.simple_date)
        with self.assertWarns(DeprecationWarning) as warnings_cm:
            obj.simple_date = None
        self.assertIsNone(obj.simple_date)

        _, _, this_module = __name__.rpartition(".")
        self.assertIn(this_module, warnings_cm.filename)
        self.assertIn(
            "None will no longer be accepted",
            str(warnings_cm.warning),
        )

    def test_assign_none_with_allow_none_false(self):
        obj = HasDateTraits(none_prohibited=UNIX_EPOCH)
        with self.assertRaises(TraitError) as exception_context:
            obj.none_prohibited = None
        message = str(exception_context.exception)
        self.assertIn("must be a date, but", message)

    def test_assign_none_with_allow_none_true(self):
        obj = HasDateTraits(none_allowed=UNIX_EPOCH)
        self.assertIsNotNone(obj.none_allowed)
        obj.none_allowed = None
        self.assertIsNone(obj.none_allowed)

    def test_assign_datetime_with_allow_datetime_false(self):
        test_datetime = datetime.datetime(1975, 2, 13)
        obj = HasDateTraits()
        with self.assertRaises(TraitError) as exception_context:
            obj.datetime_prohibited = test_datetime
        message = str(exception_context.exception)
        self.assertIn("must be a non-datetime date or None, but", message)

    def test_assign_datetime_with_allow_datetime_true(self):
        test_datetime = datetime.datetime(1975, 2, 13)
        obj = HasDateTraits()
        obj.datetime_allowed = test_datetime
        self.assertEqual(obj.datetime_allowed, test_datetime)

    def test_assign_datetime_with_allow_datetime_not_given(self):
        # For traits where "allow_datetime" is not specified, a
        # DeprecationWarning should be issued on assignment of datetime.
        test_datetime = datetime.datetime(1975, 2, 13)
        obj = HasDateTraits()
        with self.assertWarns(DeprecationWarning) as warnings_cm:
            # Note: the warning depends on the type, not the value; this case
            # should warn even though the time component of the datetime is
            # zero.
            obj.simple_date = test_datetime
        self.assertEqual(obj.simple_date, test_datetime)

        _, _, this_module = __name__.rpartition(".")
        self.assertIn(this_module, warnings_cm.filename)
        self.assertIn(
            "datetime instances will no longer be accepted",
            str(warnings_cm.warning),
        )

    def test_allow_none_false_allow_datetime_false(self):
        obj = HasDateTraits(strict=UNIX_EPOCH)
        with self.assertRaises(TraitError) as exception_context:
            obj.strict = None
        message = str(exception_context.exception)
        self.assertIn("must be a non-datetime date, but", message)

    @requires_traitsui
    def test_get_editor(self):
        obj = HasDateTraits()
        trait = obj.base_trait("epoch")
        editor_factory = trait.get_editor()
        self.assertIsInstance(editor_factory, traitsui.api.DateEditor)
